from typing import Any as _Any, override as _override

from leads.constant import SystemLiteral, ESCMode
from leads.context import Context
from leads.event import InterventionEvent, InterventionExitEvent
from leads.plugin.plugin import ESCPlugin

# (absolute, percentage)
_CALIBRATIONS: dict[ESCMode, tuple[float | None, float | None]] = {
    ESCMode.STANDARD: (1, .05),
    ESCMode.AGGRESSIVE: (4, .15),
    ESCMode.SPORT: (8, None),
    ESCMode.OFF: (None, None)
}


def do_dtcs(context: Context,
            front_wheel_speed: float,
            rear_wheel_speed: float) -> InterventionEvent:
    if ESCPlugin.adjudicate(rear_wheel_speed - front_wheel_speed, front_wheel_speed,
                            *_CALIBRATIONS[context.esc_mode()]):
        d = context.data()
        d.throttle = 0
        return InterventionEvent(context, SystemLiteral.DTCS, front_wheel_speed, rear_wheel_speed)
    return InterventionExitEvent(context, SystemLiteral.DTCS, front_wheel_speed, rear_wheel_speed)


class DTCS(ESCPlugin):
    def __init__(self) -> None:
        super(DTCS, self).__init__(("front_wheel_speed", "rear_wheel_speed"))

    @_override
    def pre_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        context.intervene(do_dtcs(context, kwargs["front_wheel_speed"], kwargs["rear_wheel_speed"]))



from json import dumps as _dumps
from typing import Any as _Any, override as _override, Literal as _Literal

from leads.data import Serializable


class ConfigTemplate(Serializable):
    def __init__(self, base: dict[str, _Any]) -> None:
        """
        All custom attributes should be public (not named after "_").
        Writable attributes should start with "w_" such as "w_debug_level".
        :param base: the base dictionary
        """
        self._d: dict[str, _Any] = base
        self._frozen: bool = False
        self.w_debug_level: _Literal["DEBUG", "INFO", "WARN", "ERROR"] = "DEBUG"
        self.data_seq_size: int = 100
        self.num_laps_timed: int = 3
        self.data_dir: str = "data"
        self.refresh()

    def __getitem__(self, name: str) -> _Any | None:
        return self.get(name)

    def __setitem__(self, name: str, value: _Any) -> None:
        self.set(name, value)

    @_override
    def __setattr__(self, name: str, value: _Any) -> None:
        if self._writable(name):
            super().__setattr__(name, value)

    @_override
    def __str__(self) -> str:
        """
        Convert to a JSON string.
        :return: the JSON string
        """
        return _dumps(self.to_dict())

    @_override
    def to_dict(self) -> dict[str, _Any]:
        return self._d.copy()

    def load(self, d: dict[str, _Any]) -> None:
        """
        Load a new dictionary and refresh.
        :param d: the dictionary to load
        """
        self._d = d
        self.refresh()

    def _writable(self, name: str) -> bool:
        """
        :param name: the configuration name
        :return: True: writable; False: readonly
        """
        return not hasattr(self, "_frozen") or not self._frozen or name.startswith("_") or name.startswith("w_")

    def set(self, name: str, value: _Any) -> None:
        """
        Set the value with the given name in the dictionary.
        :param name: the dictionary key
        :param value: the value to set
        """
        if self._writable(name):
            self._d[name] = value
            self.refresh()

    def get(self, name: str, default: _Any | None = None) -> _Any | None:
        """
        Get the value of a given name from the dictionary.
        :param name: the dictionary key
        :param default: the default value if the value does not exist
        :return: the value if it exists or else the default value
        """
        return self._d.get(name, default)

    def refresh(self) -> None:
        """
        Write the dictionary into the instance attributes.
        """
        for name in dir(self):
            if not name.startswith("_") and (v := self._d.get(name)) is not None:
                if self._writable(name):
                    setattr(self, name, v)
        self._frozen = True



from abc import ABCMeta as _ABCMeta, abstractmethod as _abstractmethod
from dataclasses import dataclass as _dataclass
from threading import Thread as _Thread
from time import sleep as _sleep
from typing import override as _override

from leads import Device as _Device
from leads_raspberry_pi.led import LED, LEDCommand
from leads_raspberry_pi.types import TransitionDirection as _TransitionDirection


class LEDGroupAnimation(object, metaclass=_ABCMeta):
    @_abstractmethod
    def do(self, command: LEDCommand, *leds: LED) -> None:
        raise NotImplementedError


class Entire(LEDGroupAnimation):
    @_override
    def do(self, command: LEDCommand, *leds: LED) -> None:
        for led in leds:
            led.write(command)


class Transition(LEDGroupAnimation):
    def __init__(self, direction: _TransitionDirection, interval: int = 1000) -> None:
        self._direction: _TransitionDirection = direction
        self._interval: float = interval * .001

    def async_do(self, command: LEDCommand, *leds: LED) -> None:
        match self._direction:
            case "left2right":
                for led in leds:
                    led.write(command)
                    _sleep(self._interval)
            case "right2left":
                for led in leds[::-1]:
                    led.write(command)
                    _sleep(self._interval)

    @_override
    def do(self, command: LEDCommand, *leds: LED) -> None:
        _Thread(name="transition", target=self.async_do, args=(command, *leds), daemon=True).start()


@_dataclass
class LEDGroupCommand(object):
    command: LEDCommand
    animation: LEDGroupAnimation


class LEDGroup(_Device):
    def __init__(self, *leds: LED) -> None:
        super().__init__()
        self._leds: tuple[LED, ...] = leds

    @_override
    def write(self, payload: LEDGroupCommand) -> None:
        payload.animation.do(payload.command, *self._leds)



from abc import abstractmethod as _abstractmethod, ABCMeta as _ABCMeta
from socket import socket as _socket, AF_INET as _AF_INET, SOCK_STREAM as _SOCK_STREAM, \
    gethostbyname_ex as _gethostbyname_ex, gethostname as _gethostname, gaierror as _gaierror, herror as _herror
from threading import Lock as _Lock, Thread as _Thread
from typing import Self as _Self, Callable as _Callable, override as _override

from leads.callback import CallbackChain
from leads.os import _thread_flags


def my_ip_addresses() -> list[str]:
    try:
        return [ip for ip in _gethostbyname_ex(_gethostname())[2] if not ip.startswith("127.")]
    except (_gaierror, _herror):
        return []


class Service(metaclass=_ABCMeta):
    def __init__(self, port: int) -> None:
        """
        :param port: the port on which the service listens
        """
        self._lock: _Lock = _Lock()
        self._port: int = port
        self._socket: _socket = _socket(_AF_INET, _SOCK_STREAM, proto=0)
        self._main_thread: _Thread | None = None

    def port(self) -> int:
        """
        Get the port that the service listens on or connects to.
        :return:
        """
        return self._port

    @_abstractmethod
    def run(self, *args, **kwargs) -> None:  # real signature unknown
        """
        Override this method to define the specific workflow.
        :param args: args
        :param kwargs: kwargs
        """
        raise NotImplementedError

    def _run(self, *args, **kwargs) -> None:
        """
        This method is equivalent to `run()`. It leaves a middle layer for possible features in subclasses.
        :param args: args passed to `run()`
        :param kwargs: kwargs passed to `run()`
        """
        self.run(*args, **kwargs)
        self.close()

    def _register_process(self, *args, **kwargs) -> None:
        """
        Register the multithread worker.
        :param args: args passed to `run()`
        :param kwargs: kwargs passed to `run()`
        """
        self._lock.acquire()
        if self._main_thread:
            raise RuntimeWarning("A service can only run once")
        try:
            self._main_thread = _Thread(name=f"service{hash(self)}", target=self._run, daemon=True, args=args,
                                        kwargs=kwargs)
        finally:
            self._lock.release()

    def _parallel_run(self, *args, **kwargs) -> None:
        """
        This method is similar to `Service._run()` except that it runs the workflow in a child thread.
        :param args: args passed to `run()`
        :param kwargs: kwargs passed to `run()`
        """
        self._register_process(*args, **kwargs)
        self._main_thread.start()

    def start(self, parallel: bool = False, *args, **kwargs) -> _Self:
        """
        This is the publicly exposed interface to start the service.
        :param parallel: True: run in a separate thread; False: run in the caller thread
        :param args: args passed to `run()`
        :param kwargs: kwargs passed to `run()`
        :return: self
        """
        try:
            return self
        finally:
            if parallel:
                self._parallel_run(*args, **kwargs)
            else:
                self._run(*args, **kwargs)

    @_abstractmethod
    def close(self) -> None:
        """
        Release the occupied resources.
        """
        raise NotImplementedError


class ConnectionBase(metaclass=_ABCMeta):
    def __init__(self, service: Service, remainder: bytes, separator: bytes) -> None:
        """
        :param service: the service to which it belongs
        :param remainder: the message remained from the last connection
        """
        self._service: Service = service
        self._remainder: bytes = remainder
        self._separator: bytes = separator

    def use_remainder(self) -> bytes:
        """
        Parse the remainder queue.
        :return: the first message from the remainder queue
        """
        if (i := self._remainder.find(self._separator)) < 0:
            msg = self._remainder
            self._remainder = b""
        elif i != len(self._remainder) - 1:
            msg = self._remainder[:i]
            self._remainder = self._remainder[i + 1:]
        else:
            msg = self._remainder[:-1]
            self._remainder = b""
        return msg

    def with_remainder(self, msg: bytes) -> bytes:
        """
        Parse the raw message and store the remaining part in the remainder queue.
        :param msg: the raw message
        :return: the first message
        """
        if (i := msg.find(self._separator)) != len(msg) - 1:
            self._remainder += msg[i + 1:]
            return msg[:i]
        return msg[:-1]

    @_abstractmethod
    def closed(self) -> bool:
        """
        :return: True: the socket is closed; False: the socket is active
        """
        raise NotImplementedError

    @_abstractmethod
    def receive(self) -> bytes | None:
        """
        :return: the message or None
        """
        raise NotImplementedError

    @_abstractmethod
    def send(self, msg: bytes) -> None:
        """
        :param msg: the message
        """
        raise NotImplementedError

    def disconnect(self) -> None:
        """
        Request disconnection.
        """
        self.send(b"disconnect")

    @_abstractmethod
    def close(self) -> None:
        """
        Directly close the socket.
        """
        raise NotImplementedError


class Connection(ConnectionBase):
    def __init__(self, service: Service, socket: _socket, address: tuple[str, int], remainder: bytes = b"",
                 separator: bytes = b";", on_close: _Callable[[_Self], None] = lambda _: None) -> None:
        super().__init__(service, remainder, separator)
        self._socket: _socket = socket
        self._address: tuple[str, int] = address
        self._on_close: _Callable[[Connection], None] = on_close

    @_override
    def __str__(self) -> str:
        """
        :return: "{address}:{port}"
        """
        return f"{self._address[0]}:{self._address[1]}"

    @_override
    def closed(self) -> bool:
        """
        Return the status of the connection.
        :return: True: closed; False: active
        """
        return self._socket.fileno() == -1

    def _require_open_socket(self, mandatory: bool = True) -> _socket:
        """
        Check if the socket is active and return it.
        :param mandatory: True: an open socket is required; False: a closed socket is acceptable
        :return: the socket object
        """
        if mandatory and self.closed():
            raise IOError("An open socket is required")
        return self._socket

    @_override
    def receive(self, chunk_size: int = 512) -> bytes | None:
        """
        Receive a full sentence from the socket.
        :param chunk_size: chunk buffer size
        :return: bytes: the message; None: failed to read (will lead to disconnection)
        """
        if self._remainder != b"":
            return self.use_remainder()
        try:
            msg = chunk = b""
            while self._separator not in chunk:
                msg += (chunk := self._require_open_socket().recv(chunk_size))
            return self.with_remainder(msg)
        except IOError:
            return

    @_override
    def send(self, msg: bytes) -> None:
        """
        Send the message to the peer.
        :param msg: the message to send
        """
        self._require_open_socket().send(msg + self._separator)
        if msg == b"disconnect":
            self.close()

    @_override
    def close(self) -> None:
        """
        Close the connection.
        """
        self._on_close(self)
        self._require_open_socket(False).close()


class Callback(CallbackChain):
    def on_initialize(self, service: Service) -> None: ...

    def on_fail(self, service: Service, error: Exception) -> None: ...

    def on_connect(self, service: Service, connection: ConnectionBase) -> None: ...

    def on_receive(self, service: Service, msg: bytes) -> None: ...

    def on_disconnect(self, service: Service, connection: ConnectionBase) -> None: ...


class Entity(Service, metaclass=_ABCMeta):
    def __init__(self, port: int, callback: Callback) -> None:
        """
        :param port: the port on which the service listens
        :param callback: the callback interface
        """
        super().__init__(port)
        self._callback: Callback = callback

    def set_callback(self, callback: Callback) -> None:
        """
        :param callback: the callback methods
        """
        callback.bind_chain(self._callback)
        self._callback = callback

    def _stage(self, connection: ConnectionBase) -> None:
        """
        Stage the connection. It loops and blocks to listen for income messages.
        :param connection: the connection to stage
        """
        while _thread_flags.active:
            msg = connection.receive()
            if msg is None or msg == b"disconnect":
                self._callback.on_disconnect(self, connection)
                return connection.close()
            self._callback.on_receive(self, msg)

    @_override
    def _run(self, *args, **kwargs) -> None:
        """
        This handles any exception raised by `super()._run()` and call the callback method `on_fail()`.
        :param args: args passed to `run()`
        :param kwargs: kwargs passed to `run()`
        """
        try:
            return super()._run(*args, **kwargs)
        except Exception as e:
            self._callback.on_fail(self, e)



from typing import override as _override

from leads import Device as _Device


class VoltageSensor(_Device):
    """
    See LEADS-Arduino.

    Supports:
    - Any analog voltage sensor
    """

    def __init__(self) -> None:
        super().__init__()
        self._voltage: float = 0

    @_override
    def update(self, data: str) -> None:
        if data.startswith(self._tag):
            self._voltage = float(data[data.find(":") + 1:])

    @_override
    def read(self) -> float:
        """
        :return: voltage
        """
        return self._voltage



from typing import Sequence as _Sequence, Any as _Any, SupportsFloat as _SupportsFloat

from leads.data_persistence.analyzer import utils as _utils
from .._computational import array as _array, ndarray as _ndarray


class DynamicProcessor(object):
    def __init__(self, data_seq: _Sequence[dict[str, _Any]]) -> None:
        self._data_seq: _Sequence[dict[str, _Any]] = data_seq

    def to_tensor(self, channels: tuple[str, ...] = ("time", "speed", "latitude", "longitude")) -> _ndarray:
        r = []
        for row in self._data_seq:
            r_row = []
            for channel in channels:
                d = row[channel]
                if not isinstance(d, _SupportsFloat):
                    raise TypeError(f"{d} ({channel}) is not a float and cannot be converted to a float")
                if getattr(_utils, f"{channel}_invalid", lambda _: False)(d):
                    raise ValueError(f"Invalid value for {channel} ({d}) at row {len(r)}")
                r_row.append(d)
            r.append(r_row)
        return _array(r)



from enum import IntEnum as _IntEnum
from typing import override as _override

from gpiozero import LED as _LED

from leads import Device as _Device


class LEDCommand(_IntEnum):
    OFF: int = 0
    ON: int = 1
    BLINK: int = 2
    BLINK_ONCE: int = 3


class LED(_Device):
    """
    Control a LED digitally.
    Supports:
    - Any 2-pin LED
    """
    def __init__(self, pin: int, blink_time_on: float = 1, blink_time_off: float = 1) -> None:
        super().__init__(pin)
        self._led: _LED = _LED(pin)
        self._blink_time_on: float = blink_time_on
        self._blink_time_off: float = blink_time_off

    @_override
    def read(self) -> bool:
        return self._led.is_active

    @_override
    def write(self, payload: LEDCommand) -> None:
        match payload:
            case LEDCommand.OFF:
                self._led.off()
            case LEDCommand.ON:
                self._led.on()
            case LEDCommand.BLINK:
                self._led.blink(self._blink_time_on, self._blink_time_off)
            case LEDCommand.BLINK_ONCE:
                self._led.blink(self._blink_time_on, self._blink_time_off, 1)



from abc import ABCMeta as _ABCMeta, abstractmethod as _abstractmethod

from leads.logger import L
from .._computational import ndarray as _ndarray


class JarvisBackend(object, metaclass=_ABCMeta):
    @_abstractmethod
    def predict(self, x: _ndarray, visual: _ndarray) -> tuple[float, float, float]:
        """
        :param x: the input data
        :param visual: the visual data
        :return: [throttle, brake, steer]
        """
        raise NotImplementedError


class Jarvis(object):
    def __init__(self, backend: JarvisBackend) -> None:
        self.backend: JarvisBackend = backend
        L.info("Jarvis is enabled\n"
               "Before proceeding, you must read the following Terms and Conditions carefully:\n"
               "Jarvis is a deep learning model that aims to provide the optimal operation based on limited inputs.\n"
               "The user must be aware of the limitations of Jarvis and always drive with caution.\n"
               "The user must pay attention to the road at all times and follow the actual situation.\n"
               "The user must not blindly trust the instructions given by Jarvis.\n"
               "Local laws and regulations unconditionally take precedence over instructions given by Jarvis.\n"
               "Any personal injury caused by erroneous instructions shall be borne by the user.")



from typing import Any as _Any, Literal as _Literal

from leads import ConfigTemplate as _ConfigTemplate, L as _L, get_config as _get_config


class Config(_ConfigTemplate):
    def __init__(self, base: dict[str, _Any]) -> None:
        self.width: int = 720
        self.height: int = 480
        self.fullscreen: bool = False
        self.no_title_bar: bool = False
        self.theme_mode: _Literal["system", "light", "dark"] = "system"
        self.manual_mode: bool = False
        self.refresh_rate: int = 30
        self.m_ratio: float = .7
        self.font_size_small: int = 14
        self.font_size_medium: int = 28
        self.font_size_large: int = 42
        self.font_size_x_large: int = 56
        self.comm_addr: str = "127.0.0.1"
        self.comm_port: int = 16900
        self.save_data: bool = False
        super().__init__(base)

    def magnify_font_sizes(self, factor: float) -> None:
        """
        Magnify font sizes by the factor.
        :param factor: the factor
        """
        if _get_config():
            raise RuntimeError("This method must be called before the config is registered")
        self._frozen = False
        self.font_size_small = int(self.font_size_small * factor)
        self.font_size_medium = int(self.font_size_medium * factor)
        self.font_size_large = int(self.font_size_large * factor)
        self.font_size_x_large = int(self.font_size_x_large * factor)
        self._frozen = True
        _L.debug(f"Font sizes magnified by {factor}")

    def auto_magnify_font_sizes(self) -> None:
        """
        Automatically magnify font sizes to match the original proportion.
        """
        self.magnify_font_sizes(self.width / Config({}).width)



from importlib.util import find_spec as _find_spec

if not _find_spec("serial"):
    raise ImportError("Please install `pyserial` to run this module\n>>>pip install pyserial")

from leads_arduino.arduino_proto import *
from leads_arduino.arduino_nano import *
from leads_arduino.arduino_micro import *
from leads_arduino.voltage_sensor import *
from leads_arduino.wheel_speed_sensor import *



from abc import ABCMeta as _ABCMeta, abstractmethod as _abstractmethod
from json import dumps as _dumps
from time import time as _time
from tkinter import Misc as _Misc, Event as _Event
from typing import Callable as _Callable, Self as _Self, TypeVar as _TypeVar, Generic as _Generic, Any as _Any, \
    Literal as _Literal

from PIL import ImageTk as _ImageTk
from customtkinter import CTk as _CTk, CTkCanvas as _CTkCanvas, get_appearance_mode as _get_appearance_mode, \
    ThemeManager as _ThemeManager, Variable as _Variable, ScalingTracker as _ScalingTracker, \
    set_appearance_mode as _set_appearance_mode
from numpy import lcm as _lcm

from leads import require_config as _require_config, DataContainer as _DataContainer
from leads.comm import Server as _Server
from leads_gui.performance_checker import PerformanceChecker
from leads_gui.system import _ASSETS_PATH
from leads_gui.types import Widget as _Widget, Color as _Color, Font as _Font


def parse_color(color: _Color) -> str:
    c = color if isinstance(color, str) else color[0] if _get_appearance_mode() == "Light" else color[1]
    return parse_color(_ThemeManager.theme["CTk"]["fg_color"]) if c == "transparent" else c


def autoscale(master: _Misc, s: float) -> float:
    return s * _ScalingTracker.get_widget_scaling(master)


def autoscale_font(master: _Misc, font: _Font) -> _Font:
    return font[0], int(autoscale(master, font[1]))


class CanvasBased(_CTkCanvas):
    def __init__(self,
                 master: _Misc,
                 theme_key: str,
                 width: float = 0,
                 height: float = 0,
                 fg_color: _Color | None = None,
                 hover_color: _Color | None = None,
                 bg_color: _Color | None = None,
                 corner_radius: float | None = None,
                 clickable: bool = False,
                 command: _Callable[[_Event], None] = lambda _: None) -> None:
        super().__init__(master, width=autoscale(master, width * master.winfo_width() if 0 < width < 1 else width),
                         height=autoscale(master, height * master.winfo_height() if 0 < height < 1 else height),
                         highlightthickness=0, cursor="hand2" if clickable else None,
                         background=parse_color(bg_color if bg_color else "transparent"))
        self._fg_color: str = parse_color(fg_color if fg_color else _ThemeManager.theme[theme_key]["fg_color"])
        self._hovering: bool = False
        try:
            self._hover_color: str = parse_color(hover_color if hover_color else _ThemeManager.theme[theme_key][
                "hover_color"])
            if clickable:
                def hover(_) -> None:
                    self._hovering = not self._hovering
                    self.render()

                self.bind("<Enter>", hover)
                self.bind("<Leave>", hover)
        except KeyError:
            self._hover_color: str = self._fg_color
        self._corner_radius: float = autoscale(master, _ThemeManager.theme[theme_key][
            "corner_radius"] if corner_radius is None else corner_radius)
        if clickable:
            self.bind("<Button-1>", command)
        self._ratio: float | None = None
        self._last_width: float = self.meta()[0]
        self._ids: dict[str, int] = {}
        self.bind("<Configure>", lambda _: self.render())

    def lock_ratio(self, ratio: float) -> _Self:
        """
        Adjust the height according to the width to keep a constant ratio.
        :param ratio: height / width
        :return: self
        """
        self._ratio = ratio
        return self

    def meta(self) -> tuple[float, float, float, float, float]:
        """
        :return: (w, h, hc, vc, limit)
        """
        return (w := self.winfo_width()), (h := self.winfo_height()), w * .5, h * .5, min(w, h)

    def collect(self, tag: str, object_id: int) -> None:
        """
        Collect a widget and link it to the tag for recycling.
        :param tag: the widget tag
        :param object_id: the widget id returned during creation
        """
        self._ids[tag] = object_id

    def clear(self, prefix: str = "") -> None:
        """
        Recycle the widgets linked to tags that start with the specified prefix.
        :param prefix: the tag prefix
        """
        if prefix:
            for tag, object_id in self._ids.copy().items():
                if tag.startswith(prefix):
                    self.delete(object_id)
                    self._ids.pop(tag)
        else:
            self.delete(*self._ids.values())
            self._ids.clear()

    def draw_fg(self, fg_color: _Color, hover_color: _Color, corner_radius: float) -> None:
        meta = self.meta()
        w, h, r = meta[0], meta[1], corner_radius * 2
        self.collect("_fg", self.create_polygon((r, 0, r, 0, w - r, 0, w - r, 0, w, 0, w, r, w, r, w, h - r, w, h - r,
                                                 w, h, w - r, h, w - r, h, r, h, r, h, 0, h, 0, h - r, 0, h - r, 0, r,
                                                 0, r, 0, 0), smooth=True,
                                                fill=hover_color if self._hovering else fg_color))

    def dynamic_renderer(self, canvas: _Self) -> None:
        """
        The dynamic renderer should only render the content-based components.
        The dynamic renderer should only access the parameters from self, and then render on the specified canvas.
        :param canvas: the canvas to render on
        """
        ...

    def partially_render(self) -> None:
        self.dynamic_renderer(self)

    def raw_renderer(self, canvas: _Self) -> None:
        """
        The raw renderer should only access the parameters from self, and then render on the specified canvas.
        :param canvas: the canvas to render on
        """
        ...

    def render(self) -> None:
        if self._ratio and (w := self.meta()[0]) != self._last_width:
            self.configure(height=w * self._ratio)
            self._last_width = w
        self.raw_renderer(self)


class TextBased(CanvasBased):
    def __init__(self,
                 master: _Misc,
                 theme_key: str,
                 width: float = 0,
                 height: float = 0,
                 font: _Font | None = None,
                 text_color: _Color | None = None,
                 fg_color: _Color | None = None,
                 hover_color: _Color | None = None,
                 bg_color: _Color | None = None,
                 corner_radius: float | None = None,
                 clickable: bool = False,
                 command: _Callable[[_Event], None] = lambda _: None) -> None:
        super().__init__(master, theme_key, width, height, fg_color, hover_color, bg_color, corner_radius, clickable,
                         command)
        self._font: _Font = autoscale_font(master, font if font else ("Arial", _require_config().font_size_small))
        self._text_color: str = parse_color(text_color if text_color else _ThemeManager.theme[theme_key]["text_color"])


class VariableControlled(object):
    def __init__(self, variable: _Variable) -> None:
        self._variable: _Variable = variable
        self._last_value: _Any = variable.get()
        self._trace_cb_name: str | None = None

    def attach(self, callback: _Callable[[], None]) -> None:
        if self._trace_cb_name:
            raise RuntimeError("Duplicated attachment")

        def unique(_, __, ___) -> None:
            if (v := self._variable.get()) != self._last_value:
                callback()
                self._last_value = v

        self._trace_cb_name = self._variable.trace_add("write", unique)

    def detach(self) -> None:
        if self._trace_cb_name:
            self._variable.trace_remove("write", self._trace_cb_name)
            self._trace_cb_name = None


class FrequencyGenerator(object, metaclass=_ABCMeta):
    def __init__(self, period: int, loops: int = -1) -> None:
        """
        :param period: the period in milliseconds
        :param loops: the number of loops or -1 to indicate infinite loops
        """
        self._period: int = period
        self._loops: int = loops
        self._last_run: float = 0

    @_abstractmethod
    def do(self) -> None:
        raise NotImplementedError

    def attempt(self) -> bool:
        """
        Attempt to run.
        :return: `True`: active; `False`: deprecated
        """
        if not self._loops:
            return False
        if (t := _time()) - self._last_run >= self._period * .001:
            self.do()
            self._loops -= 1
            self._last_run = t
        return True


class RuntimeData(object):
    start_time: int = int(_time())
    lap_time: list[int] = []
    comm: _Server | None = None

    def comm_notify(self, d: _DataContainer | dict[str, _Any]) -> None:
        if self.comm:
            self.comm.broadcast(d.encode() if isinstance(d, _DataContainer) else _dumps(d).encode())


T = _TypeVar("T", bound=RuntimeData)


class Window(_Generic[T]):
    def __init__(self,
                 width: int,
                 height: int,
                 refresh_rate: int,
                 runtime_data: T,
                 on_refresh: _Callable[[_Self], None] = lambda _: None,
                 title: str = "LEADS",
                 fullscreen: bool = True,
                 no_title_bar: bool = True,
                 theme_mode: _Literal["system", "light", "dark"] = "system") -> None:
        self._root: _CTk = _CTk()
        self._root.title(title)
        self._root.wm_iconbitmap()
        self._root.iconphoto(True, _ImageTk.PhotoImage(file=f"{_ASSETS_PATH}/logo.png"))
        self._root.overrideredirect(no_title_bar)
        _set_appearance_mode(theme_mode)
        sw, sh = self._root.winfo_screenwidth(), self._root.winfo_screenheight()
        self._width: int = sw if fullscreen else width
        self._height: int = sh if fullscreen else height
        self._root.geometry(
            f"{self._width}x{self._height}+{int((sw - self._width) / 2)}+{int((sh - self._height) / 2)}"
        )
        self._refresh_rate: int = refresh_rate
        self._runtime_data: T = runtime_data
        self._on_refresh: _Callable[[Window], None] = on_refresh
        self._frequency_generators: dict[str, FrequencyGenerator] = {}

        self._active: bool = False
        self._performance_checker: PerformanceChecker = PerformanceChecker()
        self._last_interval: float = 0

    def root(self) -> _CTk:
        return self._root

    def width(self) -> int:
        return self._width

    def height(self) -> int:
        return self._height

    def fps(self) -> float:
        return self._performance_checker.fps()

    def net_delay(self) -> float:
        return self._performance_checker.net_delay()

    def refresh_rate(self) -> int:
        return self._refresh_rate

    def runtime_data(self) -> T:
        return self._runtime_data

    def set_on_refresh(self, on_refresh: _Callable[[_Self], None]) -> None:
        self._on_refresh = on_refresh

    def add_frequency_generator(self, tag: str, frequency_generator: FrequencyGenerator) -> None:
        self._frequency_generators[tag] = frequency_generator

    def remove_frequency_generator(self, tag: str) -> None:
        try:
            self._frequency_generators.pop(tag)
        except KeyError:
            pass

    def clear_frequency_generators(self) -> None:
        self._frequency_generators.clear()

    def active(self) -> bool:
        return self._active

    def show(self) -> None:
        self._active = True

        def wrapper() -> None:
            self._on_refresh(self)
            for tag, fg in self._frequency_generators.items():
                if not fg.attempt():
                    self.remove_frequency_generator(tag)
            self._performance_checker.record_frame(self._last_interval)
            if self._active:
                self._root.after(int((ni := self._performance_checker.next_interval()) * 1000), wrapper)
                self._last_interval = ni

        self._root.after(0, wrapper)
        self._root.mainloop()

    def kill(self) -> None:
        self._active = False
        self._root.destroy()


class ContextManager(object):
    def __init__(self, window: Window) -> None:
        self._window: Window = window
        self._widgets: dict[str, _Widget] = {}

    def __setitem__(self, key: str, widget: _Widget) -> None:
        self.set(key, widget)

    def __getitem__(self, key: str) -> _Widget:
        return self.get(key)

    def set(self, key: str, widget: _Widget) -> None:
        self._widgets[key] = widget

    def get(self, key: str) -> _Widget:
        return self._widgets[key]

    def parse_layout(self, layout: list[list[str | _Widget]]) -> list[list[_Widget]]:
        for i in range(len(layout)):
            for j in range(len(layout[i])):
                e = layout[i][j]
                if isinstance(e, str):
                    layout[i][j] = self[e]
        return layout

    def layout(self, layout: list[list[str | _Widget]]) -> None:
        layout = self.parse_layout(layout)
        root = self._window.root()
        root.grid_columnconfigure(tuple(range(t := _lcm.reduce(tuple(map(len, layout))))), weight=1)
        screen_width = root.winfo_screenwidth()
        p = int(self._window.width() * .005)
        for i in range(len(layout)):
            row = layout[i]
            length = len(row)
            for j in range(length):
                widget = row[j]
                s = int(t / length)
                widget.configure(width=screen_width)
                widget.grid(row=i, column=j * s, sticky="NSEW", columnspan=s, ipadx=p, ipady=p, padx=p, pady=p)

    def window(self) -> Window:
        return self._window

    def show(self) -> None:
        self._window.show()

    def kill(self) -> None:
        self._window.kill()



from datetime import datetime as _datetime
from time import time as _time
from typing import Callable as _Callable

from customtkinter import CTkButton as _Button, CTkLabel as _Label, DoubleVar as _DoubleVar, StringVar as _StringVar, \
    CTkSegmentedButton as _CTkSegmentedButton
from pynput.keyboard import Listener as _Listener, Key as _Key, KeyCode as _KeyCode

from leads import LEADS, SystemLiteral, require_config, register_context, DTCS, ABS, EBI, ATBS, GPSSpeedCorrection, \
    ESCMode, get_controller, MAIN_CONTROLLER, L, EventListener, DataPushedEvent, UpdateEvent, has_device, \
    GPS_RECEIVER, get_device, InterventionEvent, SuspensionEvent, Event, LEFT_INDICATOR, RIGHT_INDICATOR, SFT, \
    initialize_main, format_duration
from leads.comm import Callback, Service, start_server, create_server, my_ip_addresses
from leads_audio import DIRECTION_INDICATOR_ON, DIRECTION_INDICATOR_OFF, WARNING, CONFIRM
from leads_gui import RuntimeData, Window, GForceVar, FrequencyGenerator, Left, Color, Right, ContextManager, \
    Typography, Speedometer, ProxyCanvas, SpeedTrendMeter, GForceMeter, Stopwatch, Hazard, initialize, Battery, Brake, \
    ESC, Satellite, Motor, Speed
from leads_raspberry_pi import LEDGroupCommand, LEDCommand, Transition, Entire
from leads_vec.__version__ import __version__


class CustomRuntimeData(RuntimeData):
    m1_mode: int = 0
    control_system_switch_changed: bool = False


def make_system_switch(ctx: LEADS, system: SystemLiteral, runtime_data: RuntimeData) -> _Callable[[], None]:
    def switch() -> None:
        ctx.plugin(system).enabled(not ctx.plugin(system).enabled())
        runtime_data.control_system_switch_changed = True

    return switch


def main() -> int:
    cfg = require_config()
    ctx = LEADS(data_seq_size=cfg.data_seq_size, num_laps_timed=cfg.num_laps_timed)
    register_context(ctx)
    ctx.plugin(SystemLiteral.DTCS, DTCS())
    ctx.plugin(SystemLiteral.ABS, ABS())
    ctx.plugin(SystemLiteral.EBI, EBI())
    ctx.plugin(SystemLiteral.ATBS, ATBS())
    ctx.plugin("GPS_SPEED_CORRECTION", GPSSpeedCorrection())
    w = Window(cfg.width,
               cfg.height,
               cfg.refresh_rate,
               CustomRuntimeData(),
               fullscreen=cfg.fullscreen,
               no_title_bar=cfg.no_title_bar,
               theme_mode=cfg.theme_mode)
    root = w.root()
    root.configure(cursor="dot")
    m1 = _StringVar(root, "")
    speed = _DoubleVar(root, 0)
    voltage = _StringVar(root, "")
    speed_trend = _DoubleVar(root, 0)
    g_force = GForceVar(root, 0, 0)
    esc = _StringVar(root, "STANDARD")

    class LeftIndicator(FrequencyGenerator):
        def do(self) -> None:
            uim["left"].configure(image=Left(cfg.font_size_large, Color.RED if self._loops % 2 == 1 else None))

    class RightIndicator(FrequencyGenerator):
        def do(self) -> None:
            uim["right"].configure(image=Right(cfg.font_size_large, Color.RED if self._loops % 2 == 1 else None))

    class DirectionIndicatorSound(FrequencyGenerator):
        def do(self) -> None:
            if self._loops % 2 == 1:
                DIRECTION_INDICATOR_ON.play()
            else:
                DIRECTION_INDICATOR_OFF.play()

    def render(manager: ContextManager) -> None:
        def switch_m1_mode(_) -> None:
            w.runtime_data().m1_mode = (w.runtime_data().m1_mode + 1) % 3

        manager["m1"] = Typography(root, theme_key="CTkButton", variable=m1, clickable=True, command=switch_m1_mode,
                                   font=("Arial", cfg.font_size_small - 4)).lock_ratio(cfg.m_ratio)
        manager["m2"] = Speedometer(root, variable=speed).lock_ratio(cfg.m_ratio)
        manager["m3"] = ProxyCanvas(root, "CTkButton",
                                    Typography(root, theme_key="CTkButton", variable=voltage,
                                               font=("Arial", cfg.font_size_medium - 4)),
                                    SpeedTrendMeter(root, theme_key="CTkButton", variable=speed_trend,
                                                    font=("Arial", cfg.font_size_medium - 4)),
                                    GForceMeter(root, theme_key="CTkButton", variable=g_force,
                                                font=("Arial", cfg.font_size_medium - 4))).lock_ratio(cfg.m_ratio)

        manager["comm_status"] = _Label(root, text="COMM OFFLINE", text_color="gray",
                                        font=("Arial", cfg.font_size_small))

        i = 0
        for system in SystemLiteral:
            i += 1
            system_lower = system.lower()
            manager[f"{system_lower}_status"] = _Label(root, text=f"{system} READY", text_color="green",
                                                       font=("Arial", cfg.font_size_small))
            manager[system_lower] = _Button(root, text=f"{system} ON",
                                            command=make_system_switch(ctx, SystemLiteral(system), w.runtime_data()),
                                            font=("Arial", cfg.font_size_small))

        manager["left"] = _Button(root, text="", image=Left(cfg.font_size_large),
                                  command=lambda: ctx.left_indicator(not ctx.left_indicator()))
        manager["right"] = _Button(root, text="", image=Right(cfg.font_size_large),
                                   command=lambda: ctx.right_indicator(not ctx.right_indicator()))

        def time_lap() -> None:
            ctx.time_lap()
            CONFIRM.play()

        manager["time_lap"] = _Button(root, text="", image=Stopwatch(), command=time_lap)
        manager["hazard"] = _Button(root, text="", image=Hazard(), command=lambda: ctx.hazard(not ctx.hazard()))

        def switch_esc_mode(mode: str) -> None:
            if (esc_mode := ESCMode[mode]) < 2:
                manager["esc"].configure(selected_color="green", selected_hover_color="green")
            else:
                manager["esc"].configure(selected_color="red", selected_hover_color="red")
                WARNING.play()
            ctx.esc_mode(esc_mode)
            w.runtime_data().control_system_switch_changed = True

        manager["esc"] = _CTkSegmentedButton(root, values=["STANDARD", "AGGRESSIVE", "SPORT", "OFF"], variable=esc,
                                             command=switch_esc_mode, font=("Arial", cfg.font_size_small))

    uim = initialize(w, render, ctx, get_controller(MAIN_CONTROLLER))

    class CommCallback(Callback):
        def on_fail(self, service: Service, error: Exception) -> None:
            self.super(service=service, error=error)
            L.error(f"Comm server error: {repr(error)}")

        def on_receive(self, service: Service, msg: bytes) -> None:
            self.super(service=service, msg=msg)
            match msg:
                case b"time_lap":
                    ctx.time_lap()
                case b"hazard":
                    ctx.hazard(not ctx.hazard())

    w.runtime_data().comm = start_server(create_server(cfg.comm_port, CommCallback()), True)

    class CustomListener(EventListener):
        def pre_push(self, e: DataPushedEvent) -> None:
            self.super(e)
            d = e.data.to_dict()
            d["speed_trend"] = ctx.speed_trend()
            d["lap_times"] = ctx.lap_time_list()
            w.runtime_data().comm_notify(d)

        def on_update(self, e: UpdateEvent) -> None:
            self.super(e)
            d = e.context.data()
            if w.runtime_data().m1_mode == 0:
                lap_time_list = ctx.lap_time_list()
                m1.set(f"LAP TIMES\n\n{"No Lap Timed" if len(lap_time_list) < 1 else "\n".join(
                    map(lambda t: format_duration(t * .001), lap_time_list))}")
            elif w.runtime_data().m1_mode == 1:
                if has_device(GPS_RECEIVER):
                    gps = get_device(GPS_RECEIVER).read()
                    m1.set(f"GPS {"VALID" if d.gps_valid else "NO FIX"} - {gps[4]} {gps[5]}\n\n"
                           f"{d.gps_ground_speed:.1f} KM / H\n"
                           f"LAT {d.latitude:.5f}\nLON {d.longitude:.5f}")
                else:
                    m1.set(f"GPS {"VALID" if d.gps_valid else "NO FIX"} - !NF!\n\n"
                           f"{d.gps_ground_speed:.1f} KM / H\n"
                           f"LAT {d.latitude:.5f}\nLON {d.longitude:.5f}")
            elif w.runtime_data().m1_mode == 2:
                m1.set(f"VeC {__version__.upper()}\n\n"
                       f"{_datetime.now().strftime("%Y-%m-%d %H:%M:%S")}\n"
                       f"{format_duration(duration := _time() - w.runtime_data().start_time)}\n"
                       f"{(m := d.mileage):.1f} KM - {m * 3600 / duration:.1f} KM / H\n\n"
                       f"{cfg.refresh_rate} - {w.fps():.2f} FPS - {w.net_delay() * 1000:.1f} MS\n"
                       f"{ip[-1] if len(ip := my_ip_addresses()) > 0 else "NOT FOUND"}:{w.runtime_data().comm.port()}")
            speed.set(d.speed)
            voltage.set(f"{d.voltage:.1f} V")
            st = ctx.speed_trend()
            speed_trend.set(st)
            g_force.set((d.lateral_acceleration, d.forward_acceleration))
            if w.runtime_data().comm.num_connections() < 1:
                uim["comm_status"].configure(text="COMM OFFLINE", text_color="gray")
            else:
                uim["comm_status"].configure(text="COMM ONLINE", text_color=["black", "white"])
            if w.runtime_data().control_system_switch_changed:
                for system in SystemLiteral:
                    system_lowercase = system.lower()
                    if ctx.plugin(SystemLiteral(system)).enabled():
                        uim[system_lowercase].configure(text=f"{system} ON")
                    else:
                        uim[system_lowercase].configure(text=f"{system} OFF")
                        uim[f"{system_lowercase}_status"].configure(text=f"{system} OFF", text_color=("black", "white"))
                w.runtime_data().control_system_switch_changed = False

        def pre_intervene(self, e: InterventionEvent) -> None:
            self.super(e)
            if e.system in SystemLiteral:
                uim[f"{e.system.lower()}_status"].configure(text=f"{e.system} INTEV", text_color="red")

        def post_intervene(self, e: InterventionEvent) -> None:
            self.super(e)
            if e.system in SystemLiteral:
                uim[f"{e.system.lower()}_status"].configure(text=f"{e.system} READY", text_color="green")

        def pre_suspend(self, e: SuspensionEvent) -> None:
            self.super(e)
            if e.system in SystemLiteral:
                uim[f"{e.system.lower()}_status"].configure(text=f"{e.system} SUSPD", text_color="gray")

        def post_suspend(self, e: SuspensionEvent) -> None:
            self.super(e)
            if e.system in SystemLiteral:
                uim[f"{e.system.lower()}_status"].configure(text=f"{e.system} READY", text_color="green")

        def left_indicator(self, e: Event, state: bool) -> None:
            if has_device(LEFT_INDICATOR):
                get_device(LEFT_INDICATOR).write(LEDGroupCommand(
                    LEDCommand.BLINK, Transition("left2right", 100)
                ) if state else LEDGroupCommand(LEDCommand.OFF, Entire()))
            if state:
                w.add_frequency_generator("left_indicator", LeftIndicator(500))
                w.add_frequency_generator("direction_indicator_sound", DirectionIndicatorSound(500))
            else:
                w.remove_frequency_generator("left_indicator")
                w.remove_frequency_generator("direction_indicator_sound")
                uim["left"].configure(image=Left(cfg.font_size_large, None))

        def right_indicator(self, e: Event, state: bool) -> None:
            if has_device(RIGHT_INDICATOR):
                get_device(RIGHT_INDICATOR).write(LEDGroupCommand(
                    LEDCommand.BLINK, Transition("right2left", 100)
                ) if state else LEDGroupCommand(LEDCommand.OFF, Entire()))
            if state:
                w.add_frequency_generator("right_indicator", RightIndicator(500))
                w.add_frequency_generator("direction_indicator_sound", DirectionIndicatorSound(500))
            else:
                w.remove_frequency_generator("right_indicator")
                w.remove_frequency_generator("direction_indicator_sound")
                uim["right"].configure(image=Right(cfg.font_size_large, None))

        def hazard(self, e: Event, state: bool) -> None:
            super().hazard(e, state)
            uim["hazard"].configure(image=Hazard(color=Color.RED if state else None))

    ctx.set_event_listener(CustomListener())
    uim["battery_fault"] = _Label(root, text="")
    uim["brake_fault"] = _Label(root, text="")
    uim["esc_fault"] = _Label(root, text="")
    uim["gps_fault"] = _Label(root, text="")
    uim["motor_fault"] = _Label(root, text="")
    uim["wsc_fault"] = _Label(root, text="")

    def on_fail(e: SuspensionEvent) -> None:
        match e.system:
            case "BATT":
                uim["battery_fault"].configure(image=Battery(color=Color.RED))
            case "BRAKE":
                uim["brake_fault"].configure(image=Brake(color=Color.RED))
            case "ESC":
                uim["esc_fault"].configure(image=ESC(color=Color.RED))
            case "GPS":
                uim["gps_fault"].configure(image=Satellite(color=Color.RED))
            case "MOTOR":
                uim["motor_fault"].configure(image=Motor(color=Color.RED))
            case "WSC":
                uim["wsc_fault"].configure(image=Speed(color=Color.RED))

    SFT.on_fail = on_fail

    def on_recover(e: SuspensionEvent) -> None:
        match e.system:
            case "BATT":
                uim["battery_fault"].configure(image=None)
            case "BRAKE":
                uim["brake_fault"].configure(image=None)
            case "ESC":
                uim["esc_fault"].configure(image=None)
            case "GPS":
                uim["gps_fault"].configure(image=None)
            case "MOTOR":
                uim["motor_fault"].configure(image=None)
            case "WSC":
                uim["wsc_fault"].configure(image=None)

    SFT.on_recover = on_recover
    if cfg.manual_mode:
        layout = [
            ["m1", "m2", "m3"],
            ["left", "time_lap", "hazard", "right"],
            [_Label(root, text="MANUAL MODE"), _Label(root, text="ASSISTANCE DISABLED"), "comm_status"],
            ["battery_fault", "brake_fault", "esc_fault", "gps_fault", "motor_fault", "wsc_fault"]
        ]
        ctx.esc_mode(ESCMode.OFF)
        w.runtime_data().control_system_switch_changed = True
    else:
        layout = [
            ["m1", "m2", "m3"],
            ["left", "time_lap", "hazard", "right"],
            ["battery_fault", "brake_fault", "esc_fault", "gps_fault", "motor_fault", "wsc_fault"],
            [*map(lambda s: f"{s.lower()}_status", SystemLiteral), "comm_status"],
            list(map(lambda s: s.lower(), SystemLiteral)),
            ["esc"]
        ]
    uim.layout(layout)
    root.grid_rowconfigure(2, weight=1)
    initialize_main()

    def on_press(key: _Key | _KeyCode) -> None:
        if isinstance(key, _KeyCode):
            match key.char:
                case "1":
                    make_system_switch(ctx, SystemLiteral.DTCS, w.runtime_data())()
                case "2":
                    make_system_switch(ctx, SystemLiteral.ABS, w.runtime_data())()
                case "3":
                    make_system_switch(ctx, SystemLiteral.EBI, w.runtime_data())()
                case "4":
                    make_system_switch(ctx, SystemLiteral.ATBS, w.runtime_data())()
                case "t":
                    ctx.time_lap()
        elif key == _Key.esc:
            uim.kill()

    _Listener(on_press=on_press).start()
    uim.show()
    return 0



from leads.data_persistence.analyzer.dynamic import *
from leads.data_persistence.analyzer.inference import *
from leads.data_persistence.analyzer.jarvis import *
from leads.data_persistence.analyzer.utils import *



from tkinter import Misc as _Misc
from typing import override as _override

from PIL import ImageTk as _ImageTk
from customtkinter import Variable as _Variable, DoubleVar as _DoubleVar
from numpy import sqrt as _sqrt

from leads_gui.icons import Color, Car
from leads_gui.prototype import CanvasBased, TextBased, VariableControlled, parse_color
from leads_gui.types import Font as _Font, Color as _Color


class GForceVar(_Variable):
    def __init__(self, master: _Misc, x: float = 0, y: float = 0, name: str | None = None) -> None:
        super().__init__(master, (x, y), name)

    @_override
    def set(self, value: tuple[float, float]) -> None:
        super().set(value)

    @_override
    def get(self) -> tuple[float, float]:
        return super().get()

    def magnitude(self) -> float:
        return _sqrt((v := self.get())[0] ** 2 + v[1] ** 2)


class GForceMeter(TextBased, VariableControlled):
    def __init__(self,
                 master: _Misc,
                 theme_key: str = "CTkButton",
                 width: float = 0,
                 height: float = 0,
                 variable: GForceVar | None = None,
                 font: _Font | None = None,
                 text_color: _Color | None = None,
                 fg_color: _Color | None = None,
                 hover_color: _Color | None = None,
                 bg_color: _Color | None = None,
                 corner_radius: float | None = None) -> None:
        TextBased.__init__(self, master, theme_key, width, height, font, text_color, fg_color, hover_color, bg_color,
                           corner_radius)
        VariableControlled.__init__(self, variable if variable else GForceVar(master))
        self.attach(self.partially_render)
        self._max_magnitude: float = variable.magnitude()

    @_override
    def dynamic_renderer(self, canvas: CanvasBased) -> None:
        canvas.clear("d")
        assert isinstance(self._variable, GForceVar)
        x, y = self._variable.get()
        if (magnitude := self._variable.magnitude()) > self._max_magnitude:
            self._max_magnitude = magnitude
        w, h, hc, vc, limit = canvas.meta()
        limit *= .5
        x, y = hc + max(min(x / 29.4, .9), -.9) * limit, vc + max(min(y / 29.4, .9), -.9) * limit
        r = max(min(self._max_magnitude / 29.4, .9), -.9) * limit
        canvas.collect("d0", canvas.create_oval(hc - r, vc - r, hc + r, vc + r, outline="green", width=2))
        canvas.collect("d1", canvas.create_oval(x - 4, y - 4, x + 4, y + 4, fill=self._text_color))

    @_override
    def raw_renderer(self, canvas: CanvasBased) -> None:
        canvas.clear()
        w, h, hc, vc, limit = canvas.meta()
        canvas.draw_fg(self._fg_color, self._hover_color, self._corner_radius)
        color = parse_color(("gray74", "gray60"))
        canvas.collect("f0", canvas.create_text(hc, vc, text="G", fill=self._text_color,
                                                font=(self._font[0], int(h * .1))))
        for factor in .9, .6, .3:
            r = limit * .5 * factor
            canvas.collect(f"f1{factor}", canvas.create_oval(hc - r, vc - r, hc + r, vc + r, outline=color, width=2))
        self.dynamic_renderer(canvas)


class SpeedTrendMeter(TextBased, VariableControlled):
    def __init__(self,
                 master: _Misc,
                 theme_key: str = "CTkButton",
                 width: float = 0,
                 height: float = 0,
                 variable: _DoubleVar | None = None,
                 font: _Font | None = None,
                 text_color: _Color | None = None,
                 fg_color: _Color | None = None,
                 hover_color: _Color | None = None,
                 bg_color: _Color | None = None,
                 corner_radius: float | None = None) -> None:
        TextBased.__init__(self, master, theme_key, width, height, font, text_color, fg_color, hover_color, bg_color,
                           corner_radius)
        VariableControlled.__init__(self, variable if variable else _DoubleVar(master))
        self.attach(self.partially_render)
        self._image: _ImageTk.PhotoImage | None = None

    @_override
    def dynamic_renderer(self, canvas: CanvasBased) -> None:
        canvas.clear("d")
        v = self._variable.get()
        w, h, hc, vc, limit = canvas.meta()
        canvas.collect("d0", canvas.create_line(hc, vc, hc + max(min(v, .9), -.9) * hc, vc, fill=self._text_color,
                                                arrow="last", width=4))

    @_override
    def raw_renderer(self, canvas: CanvasBased) -> None:
        canvas.clear()
        w, h, hc, vc, limit = canvas.meta()
        canvas.draw_fg(self._fg_color, self._hover_color, self._corner_radius)
        if self._image is None:
            self._image = _ImageTk.PhotoImage(Car.load_source(Color[parse_color(("BLACK", "WHITE"))]).resize((limit,
                                                                                                              limit)))
        canvas.collect("f0", canvas.create_image(hc, vc, image=self._image))
        self.dynamic_renderer(canvas)



from typing import Any as _Any, override as _override

from leads.constant import SystemLiteral, ESCMode
from leads.context import Context
from leads.event import InterventionEvent, InterventionExitEvent
from leads.plugin.plugin import ESCPlugin

# (absolute, percentage)
_CALIBRATIONS: dict[ESCMode, tuple[float | None, float | None]] = {
    ESCMode.STANDARD: (.01, .001),
    ESCMode.AGGRESSIVE: (1, .01),
    ESCMode.SPORT: (2, None),
    ESCMode.OFF: (None, None)
}


def do_abs(context: Context,
           front_wheel_speed: float,
           rear_wheel_speed: float) -> InterventionEvent:
    if ESCPlugin.adjudicate(front_wheel_speed - rear_wheel_speed, rear_wheel_speed, *_CALIBRATIONS[context.esc_mode()]):
        d = context.data()
        d.brake = 0
        return InterventionEvent(context, SystemLiteral.ABS, front_wheel_speed, rear_wheel_speed)
    return InterventionExitEvent(context, SystemLiteral.ABS, front_wheel_speed, rear_wheel_speed)


class ABS(ESCPlugin):
    def __init__(self) -> None:
        super(ABS, self).__init__(("front_wheel_speed", "rear_wheel_speed"))

    @_override
    def pre_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        context.intervene(do_abs(context, kwargs["front_wheel_speed"], kwargs["rear_wheel_speed"]))



from importlib.util import find_spec as _find_spec

if not _find_spec("PIL"):
    raise ImportError("Please install `Pillow` to run this module\n>>>pip install Pillow")
if not _find_spec("customtkinter"):
    raise ImportError("Please install `customtkinter` to run this module\n>>>pip install customtkinter")

from os.path import abspath as _abspath
from typing import Callable as _Callable, Any as _Any
from customtkinter import set_default_color_theme as _set_default_color_theme

from leads import LEADS as _LEADS, Controller as _Controller
from leads_gui.prototype import *
from leads_gui.config import *
from leads_gui.icons import *
from leads_gui.accelerometer import *
from leads_gui.speedometer import *
from leads_gui.typography import *
from leads_gui.proxy import *
from leads_gui.performance_checker import *

_set_default_color_theme(f"{_abspath(__file__)[:-11]}assets/leads-theme.json")


def initialize(window: Window,
               render: _Callable[[ContextManager], None],
               leads: _LEADS[_Any],
               main_controller: _Controller) -> ContextManager:
    ctx = ContextManager(window)
    render(ctx)

    def on_refresh(_) -> None:
        leads.push(main_controller.read())
        leads.update()

    window.set_on_refresh(on_refresh)
    return ctx



from abc import abstractmethod as _abstractmethod, ABCMeta as _ABCMeta
from threading import Thread as _Thread
from typing import Any as _Any, override as _override, overload as _overload

from leads.os import _thread_flags


class Device(object):
    @_overload
    def __init__(self, *args, **kwargs) -> None:  # real signature unknown
        raise NotImplementedError

    def __init__(self, *pins: int | str) -> None:
        self._tag: str = ""
        self._parent_tags: tuple[str, ...] = ()
        self._pins: tuple[int | str, ...] = pins

    @_override
    def __str__(self) -> str:
        return f"{len(self._parent_tags)}.{self._tag}"

    def level(self) -> int:
        """
        Get the level of the device in the device tree.
        :return: the number of parents
        """
        return len(self._parent_tags)

    def tag(self, tag: str | None = None) -> str | None:
        """
        Set or get the tag of the device.
        :param tag: the tag or None if getter mode
        :return: the tag or None if setter mode
        """
        if tag is None:
            return self._tag
        self._tag = tag

    def parent_tags(self) -> tuple[str, ...]:
        """
        Get the parent tags of the device.
        :return: the parent tags
        """
        return self._parent_tags[:]

    def pins_check(self, required_num: int) -> None:
        if len(self._pins) != required_num:
            raise ValueError(f"`{self.__class__.__name__}` only takes in {required_num} pins")

    def initialize(self, *parent_tags: str) -> None:
        self._parent_tags = parent_tags

    def read(self) -> _Any:
        raise NotImplementedError

    def write(self, payload: _Any) -> None:
        raise NotImplementedError

    def update(self, data: _Any) -> None:
        raise NotImplementedError

    def close(self) -> None:
        ...


class ShadowDevice(Device, metaclass=_ABCMeta):
    def __init__(self, *pins: int | str) -> None:
        super().__init__(*pins)
        self._shadow_thread: _Thread | None = None

    @_abstractmethod
    def loop(self) -> None:
        raise NotImplementedError

    def run(self) -> None:
        while _thread_flags.active:
            self.loop()

    @_override
    def initialize(self, *parent_tags: str) -> None:
        super().initialize(*parent_tags)
        self._shadow_thread = _Thread(name=f"{id(self)} shadow", target=self.run, daemon=True)
        self._shadow_thread.start()



from tkinter import Misc as _Misc
from typing import override as _override

from leads_gui.prototype import CanvasBased, VariableControlled
from leads_gui.types import Color as _Color


class ProxyCanvas(CanvasBased):
    def __init__(self,
                 master: _Misc,
                 theme_key: str,
                 *canvases: CanvasBased,
                 mode: int = 0,
                 width: float = 0,
                 height: float = 0,
                 fg_color: _Color | None = None,
                 hover_color: _Color | None = None,
                 bg_color: _Color | None = None,
                 corner_radius: float | None = None) -> None:
        super().__init__(master, theme_key, width, height, fg_color, hover_color, bg_color, corner_radius, True,
                         lambda _: self.next_mode())
        for canvas in canvases:
            if isinstance(canvas, VariableControlled):
                canvas.detach()
        self._canvases: tuple[CanvasBased, ...] = canvases
        self._mode: int = mode
        self._attach()

    def _attach(self) -> None:
        if isinstance(canvas := self._canvases[self._mode], VariableControlled):
            canvas.attach(self.partially_render)

    def next_mode(self) -> None:
        self.mode((self._mode + 1) % len(self._canvases))

    def mode(self, mode: int | None = None) -> int | None:
        if self._mode is None:
            return self._mode
        if isinstance(canvas := self._canvases[self._mode], VariableControlled):
            canvas.detach()
        self._mode = mode
        self.render()
        self._attach()

    @_override
    def dynamic_renderer(self, canvas: CanvasBased) -> None:
        self._canvases[self._mode].dynamic_renderer(canvas)

    @_override
    def raw_renderer(self, canvas: CanvasBased) -> None:
        self._canvases[self._mode].raw_renderer(canvas)



from abc import ABCMeta as _ABCMeta, abstractmethod as _abstractmethod
from collections import deque as _deque
from time import time as _time
from typing import TypeVar as _TypeVar, Generic as _Generic

from numpy import diff as _diff, average as _average, array as _array

from leads.constant import ESCMode
from leads.data import DataContainer

T = _TypeVar("T", bound=DataContainer)


def _check_data_type(data: T, superclass: type[DataContainer] = DataContainer) -> None:
    if not isinstance(data, superclass):
        raise TypeError(f"New data must inherit from `{superclass}`")


class Context(_Generic[T], metaclass=_ABCMeta):
    def __init__(self, initial_data: T | None, data_seq_size: int, num_laps_timed: int) -> None:
        """
        :param initial_data: initial data
        :param data_seq_size: buffer size of history data
        :param num_laps_timed: number of timed laps retained
        """
        if initial_data:
            _check_data_type(initial_data, DataContainer)
        else:
            initial_data = DataContainer()
        self._initial_data_type: type[DataContainer] = type(initial_data)
        if data_seq_size < 1:
            raise ValueError("`data_seq_size` must be greater or equal to 1")
        self._data_seq: _deque[DataContainer] = _deque((initial_data,), maxlen=data_seq_size)
        self._speed_seq: _deque[float] = _deque(maxlen=data_seq_size)
        self._lap_time_seq: _deque[int] = _deque((int(_time() * 1000),), maxlen=num_laps_timed + 1)
        self._esc_mode: ESCMode = ESCMode.STANDARD
        self._left_indicator: bool = False
        self._right_indicator: bool = False
        self._hazard: bool = False

    def data(self) -> T:
        """
        :return: a copy of the current data container
        """
        return self._data_seq[-1]

    def push(self, data: T) -> None:
        """
        Push new data into the sequence.
        :param data: the new data
        """
        _check_data_type(data, self._initial_data_type)
        self._data_seq.append(data)
        self._speed_seq.append(data.speed)

    def esc_mode(self, esc_mode: ESCMode | None = None) -> ESCMode | None:
        """
        Set or get the ESC mode.
        :param esc_mode: the ESC mode or None if getter mode
        :return: the ESC mode or None if setter mode
        """
        if esc_mode is None:
            return self._esc_mode
        self._esc_mode = esc_mode

    @_abstractmethod
    def update(self) -> None:
        raise NotImplementedError

    @_abstractmethod
    def intervene(self, *args, **kwargs) -> None:  # real signature unknown
        raise NotImplementedError

    @_abstractmethod
    def suspend(self, *args, **kwargs) -> None:  # real signature unknown
        raise NotImplementedError

    def time_lap(self) -> None:
        self._lap_time_seq.append(int(_time() * 1000))

    def lap_time_list(self) -> list[int]:
        return [self._lap_time_seq[i] - self._lap_time_seq[i - 1] for i in range(1, len(self._lap_time_seq))]

    def speed_trend(self) -> float:
        return float(_average(_diff(_array(self._speed_seq)))) if len(self._speed_seq) > 1 else 0

    def left_indicator(self, left_indicator: bool | None = None, override: bool = False) -> bool | None:
        if self._hazard and not override:
            return True
        if left_indicator is None:
            return self._left_indicator
        self._left_indicator = left_indicator

    def right_indicator(self, right_indicator: bool | None = None, override: bool = False) -> bool | None:
        if self._hazard and not override:
            return True
        if right_indicator is None:
            return self._right_indicator
        self._right_indicator = right_indicator

    def hazard(self, hazard: bool | None = None) -> bool | None:
        """
        Set or get the hazard light status.
        :param hazard: True: hazard light on; False: hazard light off; None: getter mode
        :return: the hazard light status or None if setter mode
        """
        if hazard is None:
            return self._hazard
        self.left_indicator(hazard, True)
        self.right_indicator(hazard, True)
        self._hazard = hazard



from typing import override as _override

from leads import device, controller, MAIN_CONTROLLER, LEFT_FRONT_WHEEL_SPEED_SENSOR, RIGHT_FRONT_WHEEL_SPEED_SENSOR, \
    Controller, CENTER_REAR_WHEEL_SPEED_SENSOR, require_config, mark_device, ODOMETER, GPS_RECEIVER, \
    ConcurrentOdometer, LEFT_INDICATOR, RIGHT_INDICATOR, VOLTAGE_SENSOR, DataContainer
from leads_arduino import ArduinoMicro, WheelSpeedSensor, VoltageSensor
from leads_gui import Config
from leads_raspberry_pi import NMEAGPSReceiver, LEDGroup, LED, LEDGroupCommand, LEDCommand, Entire

config: Config = require_config()
GPS_ONLY: int = config.get("gps_only", False)
BAUD_RATE: int = config.get("baud_rate", 9600)
POWER_CONTROLLER_PORT: str = config.get("power_controller_port", "COM4")
WHEEL_SPEED_CONTROLLER_PORT: str = config.get("wheel_speed_controller_port", "COM3")
GPS_RECEIVER_PORT: str = config.get("gps_receiver_port", "COM5")
FRONT_WHEEL_DIAMETER: float = config.get("front_wheel_diameter", 20)  # 20 inches
REAR_WHEEL_DIAMETER: float = config.get("rear_wheel_diameter", 20)  # 20 inches
NUM_DIVISIONS: int = config.get("num_divisions", 1)
THROTTLE_PEDAL_PIN: int = config.get("throttle_pedal_pin", 2)
BRAKE_PEDAL_PIN: int = config.get("brake_pedal_pin", 3)
VOLTAGE_SENSOR_PIN: int = config.get("voltage_sensor_pin", 4)


@controller(MAIN_CONTROLLER)
class VeCController(Controller):
    @_override
    def read(self) -> DataContainer:
        universal = {
            "mileage": self.device(ODOMETER).read(),
            "gps_valid": (gps := self.device(GPS_RECEIVER).read())[0],
            "gps_ground_speed": gps[1],
            "latitude": gps[2],
            "longitude": gps[3],
            **self.device("pc").read()
        }
        return DataContainer(**({"speed": gps[0]} if GPS_ONLY else self.device("wsc").read()), **universal)


@controller("pc", MAIN_CONTROLLER, (POWER_CONTROLLER_PORT, BAUD_RATE))
class PowerController(ArduinoMicro):
    @_override
    def initialize(self, *parent_tags: str) -> None:
        mark_device(self, "POWER", "BATT", "MOTOR")
        super().initialize(*parent_tags)

    @_override
    def read(self) -> dict[str, float]:
        return {"voltage": self.device(VOLTAGE_SENSOR).read()}

    @_override
    def write(self, payload: float) -> None:
        super().write(str(payload).encode())


@device(VOLTAGE_SENSOR, "pc")
class BatteryVoltageSensor(VoltageSensor):
    @_override
    def initialize(self, *parent_tags: str) -> None:
        mark_device(self, "POWER", "BATT")
        super().initialize(*parent_tags)


@controller("wsc", MAIN_CONTROLLER, (WHEEL_SPEED_CONTROLLER_PORT, BAUD_RATE))
class WheelSpeedController(ArduinoMicro):
    @_override
    def initialize(self, *parent_tags: str) -> None:
        mark_device(self, "WSC", "ESC")
        super().initialize(*parent_tags)

    @_override
    def read(self) -> dict[str, float]:
        lfws = self.device(LEFT_FRONT_WHEEL_SPEED_SENSOR).read()
        rfws = self.device(RIGHT_FRONT_WHEEL_SPEED_SENSOR).read()
        front_wheel_speed = (lfws + rfws) * .5
        return {"speed": min(lfws, rfws, rws := self.device(CENTER_REAR_WHEEL_SPEED_SENSOR).read()),
                "front_wheel_speed": front_wheel_speed, "rear_wheel_speed": rws}


@device(ODOMETER, MAIN_CONTROLLER)
class AverageOdometer(ConcurrentOdometer):
    @_override
    def read(self) -> float:
        return super().read() / 3


@device(*(((LEFT_FRONT_WHEEL_SPEED_SENSOR, RIGHT_FRONT_WHEEL_SPEED_SENSOR, CENTER_REAR_WHEEL_SPEED_SENSOR), "wsc",
           [(FRONT_WHEEL_DIAMETER, NUM_DIVISIONS, ODOMETER), (FRONT_WHEEL_DIAMETER, NUM_DIVISIONS, ODOMETER),
            (REAR_WHEEL_DIAMETER, NUM_DIVISIONS, ODOMETER)])))
class WheelSpeedSensors(WheelSpeedSensor):
    @_override
    def initialize(self, *parent_tags: str) -> None:
        mark_device(self, "WSC", "ESC")
        super().initialize(*parent_tags)


@device(GPS_RECEIVER, MAIN_CONTROLLER, (GPS_RECEIVER_PORT,))
class GPS(NMEAGPSReceiver):
    @_override
    def initialize(self, *parent_tags: str) -> None:
        mark_device(self, "GPS")
        super().initialize(*parent_tags)


@device((LEFT_INDICATOR, RIGHT_INDICATOR), MAIN_CONTROLLER, [
    (LED(5, .5, .5), LED(6, .5, .5), LED(26, .5, .5)), (LED(17, .5, .5), LED(27, .5, .5), LED(22, .5, .5))
])
class DirectionIndicators(LEDGroup):
    @_override
    def initialize(self, *parent_tags: str) -> None:
        mark_device(self, "DI")
        super().initialize(*parent_tags)
        self.write(LEDGroupCommand(LEDCommand.BLINK_ONCE, Entire()))



from importlib.util import find_spec as _find_spec

if not _find_spec("cv2"):
    raise ImportError("Please install `opencv-python` to run this module\n>>>pip install opencv-python")

from leads_video.camera import *



from os import chmod as _chmod, getlogin as _get_login, makedirs as _mkdirs
from os.path import abspath as _abspath, exists as _exists
from subprocess import run as _run

from leads import L as _L
from leads_gui import Config as _Config
from leads_gui.system import get_system_kernel as _get_system_kernel


def register_leads_vec() -> None:
    if _get_system_kernel() != "linux":
        raise SystemError("Unsupported operating system")
    if not _exists("/usr/local/leads/config.json"):
        _L.debug("Config file not found. Creating \"/usr/local/leads/config.json\"...")
        if not _exists("/usr/local/leads"):
            _mkdirs("/usr/local/leads")
        with open("/usr/local/leads/config.json", "w") as f:
            f.write(str(_Config({})))
    _chmod("/usr/local/leads/config.json", 0x644)
    _chmod(script := f"{_abspath(__file__)[:-10]}leads-vec.service.sh", 0o755)
    if not _exists(user_systemd := f"/home/{(username := _get_login())}/.config/systemd/user"):
        _L.debug(f"User Systemd not found. Creating \"{user_systemd}\"...")
        _mkdirs(user_systemd)
    with open(f"{user_systemd}/leads-vec.service", "w") as f:
        f.write(
            "[Unit]\n"
            "Description=LEADS VeC\n"
            "After=graphical-session.target\n"
            "[Service]\n"
            "Type=simple\n"
            f"ExecStart=/bin/bash {script}\n"
            f"Restart=always\n"
            f"RestartSec=1s\n"
            "[Install]\n"
            "WantedBy=default.target"
        )
    if not _exists(user_systemd_wants := f"{user_systemd}/default.target.wants"):
        _L.debug(f"User Systemd broken. Creating \"{user_systemd_wants}\"...")
        _mkdirs(user_systemd_wants)
    _chmod(user_systemd_wants, 0o777)
    _run(("usermod", "-aG", "dialout", username))



from typing import override as _override

from leads.comm.prototype import Entity, Connection


class Client(Entity):
    """
    You should use `create_client()` and `start_client()` instead of directly calling any method.
    """
    _connection: Connection | None = None

    @_override
    def run(self, server_address: str) -> None:
        """
        Establish a connection and stage it.
        :param server_address: the server address to which the client connects
        """
        self._callback.on_initialize(self)
        self._socket.connect((server_address, self._port))
        self._callback.on_connect(self, connection := Connection(self, self._socket, (server_address, self._port)))
        self._connection = connection
        self._stage(connection)

    def send(self, msg: bytes) -> None:
        """
        Send the message to the server.
        :param msg: the message to send
        """
        if not self._connection:
            raise IOError("Client must be running to perform this operation")
        self._connection.send(msg)

    @_override
    def close(self) -> None:
        """
        Close the connection.
        """
        if self._connection:
            self._connection.close()



from argparse import ArgumentParser as _ArgumentParser, BooleanOptionalAction as _BooleanOptionalAction
from importlib.metadata import version as _package_version, PackageNotFoundError as _PackageNotFoundError
from os import getlogin as _get_login
from os.path import abspath as _abspath
from sys import exit as _exit, version as _version
from warnings import filterwarnings as _filterwarnings

from leads import L as _L
from leads_gui.system import get_system_kernel as _get_system_kernel
from leads_vec.run import run

if __name__ == "__main__":
    _filterwarnings("ignore")
    MODULE_PATH = _abspath(__file__)[:-12]

    parser = _ArgumentParser(prog="LEADS VeC",
                             description="Lightweight Embedded Assisted Driving System VeC",
                             epilog="ProjectNeura: https://projectneura.org\n"
                                    "GitHub: https://github.com/ProjectNeura/LEADS")
    parser.add_argument("action", choices=("info", "replay", "run"))
    parser.add_argument("-c", "--config", default=None, help="specify a configuration file")
    parser.add_argument("-d", "--devices", default=f"{MODULE_PATH}/devices.py",
                        help="specify a devices module")
    parser.add_argument("-m", "--main", default=f"{MODULE_PATH}/cli.py",
                        help="specify a main module")
    parser.add_argument("-r", "--register", choices=("systemd", "config", "reverse_proxy"), default=None,
                        help="register a service")
    parser.add_argument("-t", "--theme", default=None, help="specify a theme")
    parser.add_argument("-mfs", "--magnify-font-sizes", type=float, default=1, help="magnify font sizes by a factor")
    parser.add_argument("--emu", action=_BooleanOptionalAction, default=False, help="use emulator")
    parser.add_argument("--auto-mfs", action=_BooleanOptionalAction, default=False,
                        help="automatically magnify font sizes to match the original proportion")
    parser.add_argument("--ignore-import-error", action=_BooleanOptionalAction, default=False,
                        help="ignore `ImportError`")
    args = parser.parse_args()
    if args.action == "info":
        from leads_vec.__version__ import __version__
        from ._bootloader import frpc_exists as _frpc_exists

        leads_version = "Unknown"
        try:
            leads_version = _package_version("leads")
        except _PackageNotFoundError:
            _L.warn("Failed to retrieve package version (did you install through pip?)")
        _L.info(f"LEADS VeC",
                f"System Kernel: {_get_system_kernel().upper()}",
                f"Python Version: {_version}",
                f"User: {_get_login()}",
                f"`frpc` Available: {_frpc_exists()}",
                f"Module Path: {MODULE_PATH}",
                f"LEADS Version: {leads_version}",
                f"LEADS VeC Version: {__version__}",
                sep="\n")
    else:
        if args.action == "replay":
            args.devices = f"{MODULE_PATH}/replay.py"
            args.emu = False
            _L.debug("Replay mode enabled")
        _exit(run(args.config, args.devices, args.main, args.register, args.theme, args.magnify_font_sizes, args.emu,
                  args.auto_mfs, args.ignore_import_error))
    _exit()



from tkinter import Misc as _Misc, Event as _Event
from typing import Callable as _Callable, override as _override

from customtkinter import StringVar as _StringVar

from leads_gui.prototype import CanvasBased, TextBased, VariableControlled
from leads_gui.types import Font as _Font, Color as _Color


class Typography(TextBased, VariableControlled):
    def __init__(self,
                 master: _Misc,
                 theme_key: str = "CTkLabel",
                 width: float = 0,
                 height: float = 0,
                 variable: _StringVar | None = None,
                 font: _Font | None = None,
                 text_color: _Color | None = None,
                 fg_color: _Color | None = None,
                 hover_color: _Color | None = None,
                 bg_color: _Color | None = None,
                 corner_radius: float | None = None,
                 clickable: bool = False,
                 command: _Callable[[_Event], None] = lambda _: None) -> None:
        TextBased.__init__(self, master, theme_key, width, height, font, text_color, fg_color, hover_color, bg_color,
                           corner_radius, clickable, command)
        VariableControlled.__init__(self, variable if variable else _StringVar(master))
        self.attach(self.partially_render)

    @_override
    def dynamic_renderer(self, canvas: CanvasBased) -> None:
        canvas.clear("d")
        v = self._variable.get()
        w, h, hc, vc, limit = canvas.meta()
        font = self._font
        if (target_font_size := h - 28) < font[1]:
            font = (font[0], target_font_size)
        canvas.collect("d0", canvas.create_text(w * .5, h * .5, text=v, justify="center", fill=self._text_color,
                                                font=font))

    @_override
    def raw_renderer(self, canvas: CanvasBased) -> None:
        canvas.clear()
        canvas.draw_fg(self._fg_color, self._hover_color, self._corner_radius)
        self.dynamic_renderer(canvas)



from typing import override as _override

from cv2 import VideoCapture as _VideoCapture
from numpy import ndarray as _ndarray

from leads import Device as _Device


class Camera(_Device):
    def __init__(self, port: int) -> None:
        super().__init__(port)
        self._video_capture: _VideoCapture | None = None
        self._resolution: tuple[int, int] | None = None

    @_override
    def initialize(self, *parent_tags: str) -> None:
        super().initialize(*parent_tags)
        self._video_capture = _VideoCapture(self._pins[0])

    @_override
    def write(self, payload: tuple[int, int] | None) -> None:
        self._resolution = payload

    @_override
    def read(self) -> _ndarray:
        rval, frame = self._video_capture.read()
        return frame

    @_override
    def close(self) -> None:
        self._video_capture.release()



from atexit import register
from datetime import datetime
from json import loads, JSONDecodeError
from os import makedirs
from os.path import abspath, exists
from time import sleep
from typing import Any

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from leads import require_config, L, DataContainer
from leads.comm import Service, Client, start_client, create_client, Callback, Connection, ConnectionBase
from leads.data_persistence import DataPersistence, Vector, CSV, DEFAULT_HEADER_FULL
from leads_gui import Config

config: Config = require_config()
if not exists(config.data_dir):
    L.debug(f"Data directory not found. Creating \"{abspath(config.data_dir)}\"...")
    makedirs(config.data_dir)

data_record: DataPersistence[DataContainer] = DataPersistence(1, compressor=lambda o, s: o[-s:])
time_stamp_record: DataPersistence[int] = DataPersistence(2000)
speed_record: DataPersistence[float] = DataPersistence(2000)
acceleration_record: DataPersistence[float] = DataPersistence(2000)
voltage_record: DataPersistence[float] = DataPersistence(2000)
gps_record: DataPersistence[Vector[float]] = DataPersistence(2000)
csv = CSV(f"{config.data_dir}/{datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}.csv", DEFAULT_HEADER_FULL,
          time_stamp_record, voltage_record, speed_record)


def retry(service: Service) -> Client:
    L.warn("Retrying connection...")
    return start_client(config.comm_addr, create_client(service.port(), callback), True)


class CommCallback(Callback):
    def __init__(self) -> None:
        super().__init__()
        self.client: Client = start_client(config.comm_addr, create_client(config.comm_port, self), True)

    def on_connect(self, service: Service, connection: Connection) -> None:
        self.super(service=service, connection=connection)
        L.info("Connected")

    def on_fail(self, service: Service, error: Exception) -> None:
        self.super(service=service, error=error)
        L.error(f"Comm client error: {repr(error)}")
        sleep(10)
        assert isinstance(service, Client)
        self.client = retry(service)

    def on_receive(self, service: Service, msg: bytes) -> None:
        self.super(service=service, msg=msg)
        try:
            d = loads(msg.decode())
            data_record.append(d)
            acceleration_record.append(Vector(d["forward_acceleration"], d["lateral_acceleration"]))
            gps_record.append(Vector(d["latitude"], d["longitude"]))
            if config.save_data:
                csv.write_frame(*(d[key] for key in csv.header()))
            else:
                time_stamp_record.append(int(d["t"]))
                speed_record.append(d["speed"])
                voltage_record.append(d["voltage"])
        except JSONDecodeError:
            pass

    def on_disconnect(self, service: Service, connection: ConnectionBase) -> None:
        self.super(service=service, connection=connection)
        L.info("Disconnected")
        sleep(10)
        assert isinstance(service, Client)
        self.client = retry(service)


callback: CommCallback = CommCallback()

app: FastAPI = FastAPI(title="LEADS VeC Remote Analyst")

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["GET"],
    allow_headers=["*"],
)


@app.get("/")
async def index() -> str:
    return "LEADS VeC Remote Controller"


@app.get("/current")
async def current() -> dict[str, Any]:
    return data_record[-1] if len(data_record) > 0 else DataContainer().to_dict()


@app.get("/time_stamp")
async def time_stamp() -> list[int]:
    return time_stamp_record.to_list()


@app.get("/speed")
async def speed() -> list[float]:
    return speed_record.to_list()


@app.get("/time_lap")
async def time_lap() -> str:
    callback.client.send(b"time_lap")
    return "done"


@app.get("/hazard")
async def hazard() -> str:
    callback.client.send(b"hazard")
    return "done"


register(csv.close)



from leads.comm.prototype import Callback
from leads.comm.server.server import Server


def create_server(port: int = 16900, callback: Callback = Callback()) -> Server:
    """
    Create a server service.
    :param port: the port on which the server listens
    :param callback: the callback methods
    :return: the server service
    """
    return Server(port, callback)


def start_server(target: Server = create_server(), parallel: bool = False) -> Server:
    """
    Starts the server service.
    :param target: the server service to start
    :param parallel: True: run in a separate thread; False: run in the caller thread
    :return: the server service
    """
    return target.start(parallel)



from json import dumps as _dumps
from time import time as _time
from typing import override as _override, Any as _Any

from numpy import radians as _radians, degrees as _degrees, cos as _cos


class Serializable(object):
    def to_dict(self) -> dict[str, _Any]:
        return {n: v for n in dir(self) if (v := getattr(self, n)) is not None and not callable(v) and not n.startswith(
            "_")}


class DataContainer(Serializable):
    def __init__(self,
                 voltage: float = 0,
                 speed: float = 0,
                 front_wheel_speed: float = 0,
                 rear_wheel_speed: float = 0,
                 forward_acceleration: float = 0,
                 lateral_acceleration: float = 0,
                 mileage: float = 0,
                 gps_valid: bool = False,
                 gps_ground_speed: float = 0,
                 latitude: float = 0,
                 longitude: float = 0) -> None:
        self._time_stamp: int = int(_time() * 1000)
        self.voltage: float = voltage
        self.speed: float = speed
        self.front_wheel_speed: float = front_wheel_speed
        self.rear_wheel_speed: float = rear_wheel_speed
        self.forward_acceleration: float = forward_acceleration
        self.lateral_acceleration: float = lateral_acceleration
        self.mileage: float = mileage
        self.gps_valid: bool = gps_valid
        self.gps_ground_speed: float = gps_ground_speed
        self.latitude: float = latitude
        self.longitude: float = longitude

        self.throttle: float = 0
        self.brake: float = 0

    @_override
    def __str__(self) -> str:
        return _dumps(self.to_dict())

    def reset_time_stamp(self) -> None:
        """
        Reset the data's time stamp to now.
        """
        self._time_stamp = int(_time() * 1000)

    def time_stamp(self) -> int:
        """
        Get the data's time stamp.
        :return: the time stamp (ms)
        """
        return self._time_stamp

    @_override
    def to_dict(self) -> dict[str, _Any]:
        """
        Convert the data into a dictionary.
        :return: a dictionary that contains all custom attributes of the container
        """
        (r := super().to_dict())["t"] = self._time_stamp
        return r

    def encode(self) -> bytes:
        """
        Encode the data into bytes for network transaction purposes.
        :return: JSON in bytes
        """
        return str(self).encode()


def dlat2meters(dlat: float) -> float:
    return dlat * 111320


def meters2dlat(meters: float) -> float:
    return meters / 111320


def dlon2meters(dlon: float, lat: float) -> float:
    return _radians(6378137 * dlon * _cos(_radians(lat)))


def meters2dlon(meters: float, lat: float) -> float:
    return _degrees(meters / 6378137 / _cos(_radians(lat)))


def format_duration(duration: float) -> str:
    """
    Wrap the duration into a formatted string.
    :param duration: the duration in seconds
    :return: the formatted string
    """
    return f"{(duration := int(duration)) // 60} MIN {duration % 60} SEC"



from leads import controller, MAIN_CONTROLLER, require_config, DataContainer
from leads.data_persistence import CSVDataset
from leads_emulation.replay import ReplayController


@controller(MAIN_CONTROLLER, args=(CSVDataset(f"{require_config().data_dir}/main.csv"), DataContainer))
class MainController(ReplayController[DataContainer]):
    pass



from leads.comm.client.client import Client
from leads.comm.prototype import Callback


def create_client(port: int = 16900, callback: Callback = Callback()) -> Client:
    """
    Create a client service.
    :param port: the port to which the client connects
    :param callback: the callback methods
    :return: the client service
    """
    return Client(port, callback)


def start_client(server_address: str, target: Client = create_client(), parallel: bool = False) -> Client:
    """
    Starts the client service.
    :param server_address: the server address to which the client connects
    :param target: the client service to start
    :param parallel: True: run in a separate thread; False: run in the caller thread
    :return: the client service
    """
    return target.start(parallel, server_address=server_address)



from typing import Callable as _Callable

from leads.dt import Device
from leads.event import SuspensionEvent
from leads.logger import L
from leads.registry import require_context


def mark_device(device: Device, system: str, *related: str) -> None:
    if hasattr(device, "__device_system__"):
        setattr(device, "__device_system__", getattr(device, "__device_system__") + related)
    setattr(device, "__device_system__", [system, *related])


def read_device_marker(device: Device) -> list[str] | None:
    return getattr(device, "__device_system__") if hasattr(device, "__device_system__") else None


class SystemFailureTracer(object):
    def __init__(self) -> None:
        super().__init__()
        self.on_fail: _Callable[[SuspensionEvent], None] = lambda _: None
        self.on_recover: _Callable[[SuspensionEvent], None] = lambda _: None
        self.on_device_fail: _Callable[[Device, str | Exception], None] = lambda _, __: None
        self.on_device_recover: _Callable[[Device], None] = lambda _: None
        self._system_failure: dict[str, int] = {}

    def fail(self, device: Device, error: str | Exception) -> None:
        if isinstance(error, Exception):
            error = repr(error)
        if not (systems := read_device_marker(device)):
            raise RuntimeWarning(f"No system marked for device {device}")
        self.on_device_fail(device, error)
        L.error(f"{device} error: {error}")
        for system in systems:
            if system not in self._system_failure:
                self._system_failure[system] = 0
            self._system_failure[system] += 1
            self.on_fail(e := SuspensionEvent(context := require_context(), system, error))
            context.suspend(e)

    def recover(self, device: Device) -> None:
        if not (systems := read_device_marker(device)):
            raise RuntimeWarning(f"System not marked for device {device}")
        self.on_device_recover(device)
        L.debug(f"{device} recovered")
        for system in systems:
            if system not in self._system_failure:
                continue
            self._system_failure[system] -= 1
            if self._system_failure[system] > 0:
                continue
            self._system_failure.pop(system)
            self.on_recover(SuspensionEvent(require_context(), system, "Recovered"))


SFT: SystemFailureTracer = SystemFailureTracer()



from time import time as _time
from typing import override as _override

from numpy import pi as _pi

from leads import Device as _Device, get_device as _get_device, Odometer as _Odometer


def rpm2kmh(rpm: float, wheel_circumference: float) -> float:
    """
    :param rpm: revolutions per minute
    :param wheel_circumference: wheel circumference in centimeters
    :return: speed in kilometers per hour
    """
    return rpm * wheel_circumference * .0006


class WheelSpeedSensor(_Device):
    """
    See LEADS-Arduino.

    Supports:
    - Any hall effect sensor (switch)
    """

    def __init__(self, wheel_diameter: float, num_divisions: int = 1, odometer_tag: str | None = None) -> None:
        super().__init__()
        self._wheel_circumference: float = wheel_diameter * 2.54 * _pi
        self._num_divisions: int = num_divisions
        self._wheel_speed: float = 0
        self._last_valid: float = 0
        self._odometer_tag: str | None = odometer_tag
        self._odometer: _Odometer | None = None

    @_override
    def initialize(self, *parent_tags: str) -> None:
        super().initialize(*parent_tags)
        if self._odometer_tag:
            self._odometer = _get_device(self._odometer_tag)

    @_override
    def update(self, data: str) -> None:
        if data.startswith(self._tag):
            self._wheel_speed = rpm2kmh(float(data[data.find(":") + 1:]),
                                        self._wheel_circumference) / self._num_divisions
            self._last_valid = _time()
            if self._odometer:
                self._odometer.write(self._wheel_circumference * .00001)

    @_override
    def read(self) -> float:
        """
        :return: speed in kilometers per hour
        """
        # add .0000000001 to avoid zero division
        r = rpm2kmh(60 / (.0000000001 + _time() - self._last_valid), self._wheel_circumference) / self._num_divisions
        return 0 if r < .1 else r if r < 5 else self._wheel_speed



from typing import Any as _Any

from leads.data import dlat2meters, dlon2meters
from .._computational import sqrt as _sqrt


def time_invalid(o: _Any) -> bool:
    return not isinstance(o, int)


def speed_invalid(o: _Any) -> bool:
    return not isinstance(o, int | float) or o != o or o < 0


def acceleration_invalid(o: _Any) -> bool:
    return not isinstance(o, int | float) or o != o


def mileage_invalid(o: _Any) -> bool:
    return not isinstance(o, int | float) or o != o


def latitude_invalid(o: _Any) -> bool:
    return not isinstance(o, int | float) or o != o or not -90 < o < 90


def longitude_invalid(o: _Any) -> bool:
    return not isinstance(o, int | float) or o != o or not -180 < o < 180


def distance_between(lat_0: float, lon_0: float, lat: float, lon: float) -> float:
    """
    Calculate the distance between two locations on the Earth.
    :param lat_0: the latitude of the first location
    :param lon_0: the longitude of the first location
    :param lat: the latitude of the second location
    :param lon: the longitude of the second location
    :return:
    """
    return _sqrt(dlon2meters(lon - lon_0, .5 * (lat_0 + lat)) ** 2 + dlat2meters(lat - lat_0) ** 2)



from abc import ABCMeta as _ABCMeta, abstractmethod as _abstractmethod
from typing import Any as _Any, override as _override, Generator as _Generator

from leads.data_persistence.analyzer.utils import time_invalid, speed_invalid, acceleration_invalid, \
    mileage_invalid, latitude_invalid, longitude_invalid, distance_between
from leads.data_persistence.core import CSVDataset, DEFAULT_HEADER


class Inference(object, metaclass=_ABCMeta):
    def __init__(self, required_depth: tuple[int, int] = (0, 0)) -> None:
        """
        Declare the scale of data this inference requires.
        :param required_depth: (-depth backward, depth forward)
        """
        self._required_depth: tuple[int, int] = required_depth

    def depth(self) -> tuple[int, int]:
        """
        :return: (-depth backward, depth forward)
        """
        return self._required_depth

    @_abstractmethod
    def complete(self, *rows: dict[str, _Any], backward: bool = False) -> dict[str, _Any] | None:
        """
        Infer, based on the data flow, to complete the missing columns.
        :param rows: the data flow with the length of depth backward + depth forward
        :param backward: True: globally reversed; False: regular
        :return:
        """
        raise NotImplementedError


class SpeedInferenceBase(Inference, metaclass=_ABCMeta):
    @staticmethod
    def skip(row: dict[str, _Any]) -> bool:
        return not speed_invalid(row["speed"])


class SafeSpeedInference(SpeedInferenceBase):
    """
    Infer the speed based on the front wheel speed or the rear wheel speed.

    v = min(fws, rws)
    """

    def __init__(self) -> None:
        super().__init__((0, 0))

    @_override
    def complete(self, *rows: dict[str, _Any], backward: bool = False) -> dict[str, _Any] | None:
        row = rows[0]
        if SpeedInferenceBase.skip(row):
            return
        speed = None
        if not speed_invalid(s := row["front_wheel_speed"]):
            speed = s
        if not speed_invalid(s := row["rear_wheel_speed"]) and (speed is None or s < speed):
            speed = s
        return None if speed is None else {"speed": speed}


class SpeedInferenceByAcceleration(SpeedInferenceBase):
    """
    Infer the speed based on the acceleration.

    v = ∫a(t)dt
    """

    def __init__(self) -> None:
        super().__init__((-1, 0))

    @_override
    def complete(self, *rows: dict[str, _Any], backward: bool = False) -> dict[str, _Any] | None:
        base, target = rows
        t_0, t, v_0, a_0 = base["t"], target["t"], base["speed"], base["forward_acceleration"]
        if (SpeedInferenceBase.skip(target) or time_invalid(t_0) or
                time_invalid(t) or speed_invalid(v_0) or
                acceleration_invalid(a_0)):
            return
        a = target["forward_acceleration"]
        if acceleration_invalid(a):
            a = a_0
        return {"speed": abs(v_0 + .0018 * (a_0 + a) * (t - t_0))}


class SpeedInferenceByMileage(SpeedInferenceBase):
    """
    Infer the speed based on the mileage.

    v = ds/dt
    """

    def __init__(self) -> None:
        super().__init__((-1, 0))

    @_override
    def complete(self, *rows: dict[str, _Any], backward: bool = False) -> dict[str, _Any] | None:
        base, target = rows
        t_0, t, s_0, s = base["t"], target["t"], base["mileage"], target["mileage"]
        return None if (SpeedInferenceBase.skip(target) or time_invalid(t_0) or
                        time_invalid(t) or mileage_invalid(s_0) or
                        mileage_invalid(s)) else {
            "speed": abs(3600000 * (s - s_0) / (t - t_0))
        }


class SpeedInferenceByGPSGroundSpeed(SpeedInferenceBase):
    """
    Infer the speed based on the GPS ground speed.
    """

    def __init__(self) -> None:
        super().__init__((0, 0))

    @_override
    def complete(self, *rows: dict[str, _Any], backward: bool = False) -> dict[str, _Any] | None:
        row = rows[0]
        ground_speed = row["gps_ground_speed"]
        return None if (SpeedInferenceBase.skip(row) or not row["gps_valid"] or
                        speed_invalid(ground_speed)) else {
            "speed": ground_speed
        }


class SpeedInferenceByGPSPosition(SpeedInferenceBase):
    """
    Infer the speed based on the GPS position.

    v = ds/dt
    """

    def __init__(self) -> None:
        super().__init__((-1, 0))

    @_override
    def complete(self, *rows: dict[str, _Any], backward: bool = False) -> dict[str, _Any] | None:
        base, target = rows
        t_0, t = base["t"], target["t"]
        lat_0, lat, lon_0, lon = base["latitude"], target["latitude"], base["longitude"], target["longitude"]
        return None if (SpeedInferenceBase.skip(target) or time_invalid(t_0) or
                        time_invalid(t) or not base["gps_valid"] or not target["gps_valid"] or
                        latitude_invalid(lat_0) or latitude_invalid(lat) or
                        longitude_invalid(lon_0) or longitude_invalid(lon)) else {
            "speed": abs(3600 * distance_between(lat_0, lon_0, lat, lon) / (t - t_0))
        }


class ForwardAccelerationInferenceBase(Inference, metaclass=_ABCMeta):
    @staticmethod
    def skip(row: dict[str, _Any]) -> bool:
        return not mileage_invalid(row["forward_acceleration"])


class ForwardAccelerationInferenceBySpeed(ForwardAccelerationInferenceBase):
    """
    Infer the forward acceleration based on the speed.

    a = dv/dt
    """

    def __init__(self) -> None:
        super().__init__((0, 1))

    @_override
    def complete(self, *rows: dict[str, _Any], backward: bool = False) -> dict[str, _Any] | None:
        target, base = rows
        t_0, t, v_0, v = target["t"], base["t"], target["speed"], base["speed"]
        return None if (ForwardAccelerationInferenceBase.skip(target) or time_invalid(t_0) or
                        time_invalid(t) or speed_invalid(v_0) or
                        speed_invalid(v)) else {
            "forward_acceleration": (v - v_0) / (t - t_0)
        }


class MileageInferenceBase(Inference, metaclass=_ABCMeta):
    @staticmethod
    def skip(row: dict[str, _Any]) -> bool:
        return not mileage_invalid(row["mileage"])


class MileageInferenceBySpeed(MileageInferenceBase):
    """
    Infer the mileage based on the speed.

    s = ∫v(t)dt
    """

    def __init__(self) -> None:
        super().__init__((-1, 0))

    @_override
    def complete(self, *rows: dict[str, _Any], backward: bool = False) -> dict[str, _Any] | None:
        base, target = rows
        t_0, t, v_0, s_0 = base["t"], target["t"], base["speed"], base["mileage"]
        if (MileageInferenceBase.skip(target) or time_invalid(t_0) or time_invalid(t) or
                speed_invalid(v_0) or mileage_invalid(s_0)):
            return
        v = target["speed"]
        if speed_invalid(v):
            v = v_0
        return {"mileage": s_0 + .00000125 * (v_0 + v) * (t - t_0) / 9}


class MileageInferenceByGPSPosition(MileageInferenceBase):
    """
    Infer the mileage based on the speed.

    s = s_0 + Δs
    """

    def __init__(self) -> None:
        super().__init__((-1, 0))

    @_override
    def complete(self, *rows: dict[str, _Any], backward: bool = False) -> dict[str, _Any] | None:
        base, target = rows
        s_0 = base["mileage"]
        lat_0, lat, lon_0, lon = base["latitude"], target["latitude"], base["longitude"], target["longitude"]
        return None if (MileageInferenceBase.skip(target) or mileage_invalid(s_0) or
                        not base["gps_valid"] or not target["gps_valid"] or latitude_invalid(lat_0) or
                        latitude_invalid(lat) or longitude_invalid(lon_0) or
                        longitude_invalid(lon)) else {
            "mileage": s_0 + .001 * distance_between(lat_0, lon_0, lat, lon)
        }


class InferredDataset(CSVDataset):
    def __init__(self, file: str, chunk_size: int = 100) -> None:
        super().__init__(file, chunk_size)
        self._raw_data: tuple[dict[str, _Any], ...] = ()
        self._inferred_data: list[dict[str, _Any]] = []

    @_override
    def __len__(self) -> int:
        return len(self._raw_data)

    @staticmethod
    def merge(raw: dict[str, _Any], inferred: dict[str, _Any]) -> None:
        """
        Merge the inferred data to the raw data. Overwrite if conflicts. It directly alters the raw data object.
        :param raw: the raw data
        :param inferred: the difference data
        """
        for key in inferred.keys():
            raw[key] = inferred[key]

    def _complete(self, inferences: tuple[Inference, ...], enhanced: bool, backward: bool) -> None:
        num_rows = len(self._raw_data)
        for i in range(num_rows - 1, -1, -1) if backward else range(num_rows):
            for inference in inferences:
                p, f = inference.depth()
                p, f = (i - p, i - f - 1) if backward else (i + p, i + f + 1)
                d = []
                if (-1 < p < num_rows and -1 <= f < num_rows - 1) if backward else (
                        0 <= p < num_rows and 0 < f <= num_rows):
                    for j in range(p, f, -1 if backward else 1):
                        row = self._raw_data[j]
                        if enhanced:
                            InferredDataset.merge(row, self._inferred_data[j])
                        d.append(row)
                    if (r := inference.complete(*d, backward=backward)) is not None:
                        InferredDataset.merge(self._inferred_data[i], r)

    @_override
    def load(self) -> None:
        if self._raw_data:
            return
        super().load()
        raw_data = []
        for row in super().__iter__():
            raw_data.append(row)
        self._raw_data = tuple(raw_data)
        self._inferred_data = [{} for _ in range(len(raw_data))]

    def assume_initial_zeros(self) -> None:
        row = self._raw_data[0]
        injection = {}
        if speed_invalid(row["speed"]):
            injection["speed"] = 0
        if acceleration_invalid(row["forward_acceleration"]):
            injection["forward_acceleration"] = 0
        if mileage_invalid(row["mileage"]):
            injection["mileage"] = 0
        InferredDataset.merge(row, injection)

    def complete(self, *inferences: Inference, enhanced: bool = False, assume_initial_zeros: bool = False) -> None:
        """
        Infer the missing values in the dataset.
        :param inferences: the inferences to apply
        :param enhanced: True: use inferred data to infer other data; False: use only raw data to infer other data
        :param assume_initial_zeros: True: reasonably set any missing data in the first row to zero; False: no change
        """
        if DEFAULT_HEADER in self.read_header():
            raise KeyError("Your dataset must include the default header")
        if assume_initial_zeros:
            self.assume_initial_zeros()
        self._complete(inferences, enhanced, False)
        self._complete(inferences, enhanced, True)

    @_override
    def __iter__(self) -> _Generator[dict[str, _Any], None, None]:
        for i in range(len(self._raw_data)):
            InferredDataset.merge(row := self._raw_data[i], self._inferred_data[i])
            yield row

    @_override
    def close(self) -> None:
        super().close()
        self._raw_data = ()
        self._inferred_data.clear()



from importlib.util import find_spec as _find_spec

if not _find_spec("pynput"):
    raise ImportError("Please install `pynput` to run this module\n>>>pip install pynput")

from leads_vec.run import *



from collections import deque as _deque
from time import time as _time

from numpy import average as _average, poly1d as _poly1d, polyfit as _polyfit

from leads import require_config as _require_config


class PerformanceChecker(object):
    def __init__(self) -> None:
        self._refresh_rate: int = _require_config().refresh_rate
        self._interval: float = 1 / self._refresh_rate
        self._predicted_offset: float = 0
        self._delay_seq: _deque[float] = _deque(maxlen=self._refresh_rate * 10)
        self._net_delay_seq: _deque[float] = _deque(maxlen=self._refresh_rate * 10)
        self._last_frame: float = _time()

    def fps(self) -> float:
        return 1 / _average(self._delay_seq)

    def net_delay(self) -> float:
        return _average(self._net_delay_seq)

    def record_frame(self, interval: float) -> None:
        # add .0000000001 to avoid zero division
        self._delay_seq.append(delay := .0000000001 + (t := _time()) - self._last_frame)
        self._net_delay_seq.append(delay - interval)
        mark = len(self._net_delay_seq)
        self._predicted_offset = max(min(_poly1d(_polyfit(range(mark), self._net_delay_seq, 5))(mark + 1)
                                         if mark > self._refresh_rate else 0, self._interval), 0)
        self._last_frame = t

    def next_interval(self) -> float:
        return self._interval - self._predicted_offset



from importlib.util import spec_from_file_location as _spec_from_file_location, module_from_spec as _module_from_spec
from os.path import abspath as _abspath
from os.path import exists as _exists
from typing import Literal as _Literal

from customtkinter import set_default_color_theme as _set_default_color_theme

from leads import register_controller as _register_controller, MAIN_CONTROLLER as _MAIN_CONTROLLER, \
    L as _L, load_config as _load_config, register_config as _register_config, release as _release
from leads_gui import Config as _Config


def run(config: str, devices: str, main: str, register: _Literal["systemd", "config", "reverse_proxy"], theme: str,
        magnify_font_sizes: float, emu: bool, auto_mfs: bool, ignore_import_error: bool) -> int:
    match register:
        case "systemd":
            from ._bootloader import register_leads_vec as _create_service

            _create_service()
            _L.debug("Service registered")
            _L.debug(f"Service script is located at \"{_abspath(__file__)[:-6]}_bootloader/leads-vec.service.sh\"")
        case "config":
            if _exists("config.json"):
                r = input("\"config.json\" already exists. Overwrite? (Y/n) >>>").lower()
                if r.lower() != "y":
                    _L.error("Aborted")
                    return 1
            with open("config.json", "w") as f:
                f.write(str(_Config({})))
            _L.debug("Configuration file saved to \"config.json\"")
        case "reverse_proxy":
            from ._bootloader import start_frpc as _start_frpc

            _start_frpc()
            _L.debug("`frpc` started")
    config = _load_config(config, _Config) if config else _Config({})
    _L.debug("Configuration loaded:", str(config))
    if t := theme:
        _set_default_color_theme(t)
    if (f := magnify_font_sizes) != 1:
        config.magnify_font_sizes(f)
    if auto_mfs:
        config.auto_magnify_font_sizes()
    _register_config(config)
    spec = _spec_from_file_location("main", main)
    spec.loader.exec_module(main := _module_from_spec(spec))
    try:
        main = getattr(main, "main")
    except AttributeError:
        raise ImportError(f"No main function in \"{main}\"")

    try:
        if emu:
            raise SystemError("User specifies to use emulator")
        spec = _spec_from_file_location("_", devices)
        spec.loader.exec_module(_module_from_spec(spec))
    except (ImportError, SystemError) as e:
        _L.debug(repr(e))
        if isinstance(e, ImportError):
            if ignore_import_error:
                _L.debug(f"Ignoring import error: {repr(e)}")
                return main()
            else:
                _L.warn(f"Specified devices module ({devices}) is not available, using emulation module instead...")
        _release()
        try:
            from leads_emulation import SinController as _Controller

            _register_controller(_MAIN_CONTROLLER, _Controller(0, 200, .05))
        except ImportError as e:
            _L.error(f"Emulator error: {repr(e)}")
            return 1
    return main()



from leads.config import *
from leads.context import *
from leads.data import *
from leads.dt import *
from leads.event import *
from leads.leads import *
from leads.logger import Level, L
from leads.plugin import *
from leads.registry import *
from leads.sft import SFT, mark_device, read_device_marker



from typing import override as _override, Iterator as _Iterator, Any as _Any, TypeVar as _TypeVar, \
    Generic as _Generic

from leads import Controller as _Controller, DataContainer as _DataContainer
from leads.data_persistence import CSVDataset as _CSVDataset

T = _TypeVar("T", bound=_DataContainer)


class ReplayController(_Controller, _Generic[T]):
    def __init__(self, dataset: _CSVDataset, data_container_constructor: type[T]) -> None:
        super().__init__()
        self._dataset: _CSVDataset = dataset
        self._constructor: type[T] = data_container_constructor
        self._iterator: _Iterator[dict[str, _Any]] = iter(dataset)

    @_override
    def read(self) -> T:
        try:
            d = next(self._iterator)
        except StopIteration:
            return self._constructor()
        t = d.pop("t")
        dc = self._constructor(**d)
        dc._time_stamp = t
        return dc

    @_override
    def close(self) -> None:
        self._dataset.close()



from typing import Callable as _Callable, SupportsInt as _SupportsInt, SupportsFloat as _SupportsFloat

type Number = int | float | _SupportsInt | _SupportsFloat
type Compressor[T] = _Callable[[list[T], int], list[T]]
type OnRegister[T] = _Callable[[T], None]
type OnRegisterChain[T] = _Callable[[OnRegister[T]], OnRegister[T]]



try:
    import cupy as _np
except ImportError:
    import numpy as _np
import pandas as _pandas

mean: type[_np.mean] = _np.mean
array: type[_np.array] = _np.array
ndarray: type[_np.array] = _np.ndarray
norm: type[_np.linalg.norm] = _np.linalg.norm
sqrt: type[_np.sqrt] = _np.sqrt
minimum: type[_np.min] = _np.min
maximum: type[_np.max] = _np.max

read_csv: type[_pandas.read_csv] = _pandas.read_csv
DataFrame: type[_pandas.DataFrame] = _pandas.DataFrame
TextFileReader: type[_pandas.io.parsers.TextFileReader] = _pandas.io.parsers.TextFileReader



from operator import add as _add, sub as _sub, mul as _mul, truediv as _truediv, floordiv as _floordiv, lt as _lt, \
    le as _le, gt as _gt, ge as _ge
from typing import TextIO as _TextIO, TypeVar as _TypeVar, Generic as _Generic, Sequence as _Sequence, \
    override as _override, Self as _Self, Iterator as _Iterator, Callable as _Callable, Iterable as _Iterable, \
    Generator as _Generator, Any as _Any

from leads.types import Compressor as _Compressor
from ._computational import mean as _mean, array as _array, norm as _norm, read_csv as _read_csv, \
    DataFrame as _DataFrame, TextFileReader as _TextFileReader

T = _TypeVar("T")


def mean_compressor(sequence: list[T], target_size: int) -> list[T]:
    """
    A compression method that reduces data memory usage by averaging adjacent numbers and merging them.
    :param sequence: the sequence to compress
    :param target_size: expected size
    :return: the compressed sequence
    """
    chunk_size = int(len(sequence) / target_size)
    if chunk_size < 2:
        return sequence
    r = []
    for i in range(target_size - 1):
        r.append(_mean(sequence[i * chunk_size: (i + 1) * chunk_size]))
    r.append(_mean(sequence[(target_size - 1) * chunk_size:]))
    return r


class DataPersistence(_Sequence[T], _Generic[T]):
    def __init__(self, max_size: int = -1, chunk_scale: int = 1, compressor: _Compressor[T] = mean_compressor) -> None:
        """
        :param max_size: maximum cached size
        :param chunk_scale: chunk scaling factor (compression)
        :param compressor: compressor interface
        """
        self._max_size: int = max_size
        self._chunk_scale: int = chunk_scale
        self._compressor: _Compressor[T] = compressor
        self._data: list[T] = []
        self._chunk: list[T] = []
        self._chunk_size: int = chunk_scale

    @_override
    def __len__(self) -> int:
        return len(self._data)

    @_override
    def __getitem__(self, item: int | slice) -> T | list[T]:
        return self._data[item]

    @_override
    def __str__(self) -> str:
        return str(self._data)

    def to_list(self) -> list[T]:
        return self._data.copy()

    def _push_to_data(self, element: T) -> None:
        self._data.append(element)
        if self._max_size < 2:
            return
        if len(self._data) >= self._max_size:
            self._data = self._compressor(self._data, int(len(self._data) * .5))
            self._chunk_size *= 2

    def append(self, element: T) -> None:
        if self._chunk_size == 1:
            return self._push_to_data(element)
        self._chunk.append(element)
        if len(self._chunk) >= self._chunk_size:
            for e in self._compressor(self._chunk, self._chunk_scale):
                self._push_to_data(e)
            self._chunk.clear()


E = _TypeVar("E")


class Vector(_Sequence[E], _Iterable[E], _Generic[E]):
    def __init__(self, *coordinates: E) -> None:
        self._d: int = len(coordinates)
        self._coordinates: tuple[E, ...] = coordinates

    @_override
    def __len__(self) -> int:
        return self._d

    @_override
    def __iter__(self) -> _Iterator[E]:
        return iter(self._coordinates)

    @_override
    def __getitem__(self, item: int | slice) -> _Self:
        return Vector(*self._coordinates[item])

    @_override
    def __str__(self) -> str:
        return ";".join(map(str, self._coordinates))

    @_override
    def __eq__(self, other: _Self) -> bool:
        return self._coordinates == other._coordinates

    def _check_dimension(self, other: _Self) -> None:
        if other._d != self._d:
            raise ValueError("Cannot perform this operation on two vectors of different dimensions")

    def __neg__(self) -> _Self:
        return Vector(*(-i for i in self._coordinates))

    def __abs__(self) -> _Self:
        return Vector(*(abs(i) for i in self._coordinates))

    def _operate(self, other: _Self | E, operator: _Callable[[E, E], E]) -> _Self:
        if isinstance(other, Vector):
            self._check_dimension(other)
            return Vector(*(operator(self._coordinates[i], other._coordinates[i]) for i in range(self._d)))
        return Vector(*(operator(i, other) for i in self._coordinates))

    def __add__(self, other: _Self | E) -> _Self:
        return self._operate(other, _add)

    def __sub__(self, other: _Self | E) -> _Self:
        return self._operate(other, _sub)

    def __mul__(self, other: _Self | E) -> _Self:
        return self._operate(other, _mul)

    def __truediv__(self, other: _Self | E) -> _Self:
        return self._operate(other, _truediv)

    def __floordiv__(self, other: _Self | E) -> _Self:
        return self._operate(other, _floordiv)

    def distance(self, other: _Self) -> float:
        return float(_norm(_array(self._coordinates) - _array(other._coordinates)))

    def magnitude(self) -> float:
        return self.distance(Vector(*(0,) * self._d))

    def _compare(self, other: _Self | E, comparer: _Callable[[E, E], bool]) -> bool:
        return comparer(self.magnitude(), other.magnitude() if isinstance(other, Vector) else other)

    def __lt__(self, other: _Self | E) -> bool:
        return self._compare(other, _lt)

    def __le__(self, other: _Self | E) -> bool:
        return self._compare(other, _le)

    def __gt__(self, other: _Self | E) -> bool:
        return self._compare(other, _gt)

    def __ge__(self, other: _Self | E) -> bool:
        return self._compare(other, _ge)


class CSV(object):
    def __init__(self, file: str | _TextIO, header: tuple[str, ...], *columns: DataPersistence | None) -> None:
        self._file: _TextIO = open(file, "w") if isinstance(file, str) else file
        self._d: int = len(header)
        self._header: tuple[str, ...] = header
        if (d := self._d - len(columns)) >= 0:
            columns += (None,) * d
        else:
            raise ValueError("Unmatched columns and header")
        self._columns: tuple[DataPersistence | None, ...] = columns
        self._i: int = 0
        self.write_header()

    def header(self) -> tuple[str, ...]:
        return self._header

    def write_header(self) -> None:
        header = f"{",".join(self._header)}\n"
        if not header.startswith("index"):
            header = f"index,{header}"
        self._file.write(header)

    def write_frame(self, *data: _Any) -> None:
        if len(data) != self._d:
            raise ValueError("Unmatched data and header")
        frame = {}
        for i in range(self._d):
            frame[self._header[i]] = d = data[i]
            if column := self._columns[i]:
                column.append(d)
        _DataFrame(data=frame, index=[self._i]).to_csv(self._file, mode="a", header=False)
        self._i += 1

    def close(self) -> None:
        self._file.close()


class CSVDataset(_Iterable[dict[str, _Any]]):
    def __init__(self, file: str, chunk_size: int = 100) -> None:
        self._file: str = file
        self._chunk_size: int = chunk_size
        self._csv: _TextFileReader | None = None
        self._header: tuple[str, ...] | None = None
        self._contains_index: bool = False

    def require_loaded(self) -> None:
        if not self._csv or not self._header:
            self.load()

    def read_header(self) -> tuple[str, ...]:
        self.require_loaded()
        return self._header

    @_override
    def __iter__(self) -> _Generator[dict[str, _Any], None, None]:
        self.require_loaded()
        while True:
            try:
                chunk = next(self._csv)
            except StopIteration:
                break
            for i in range(len(chunk)):
                r = chunk.iloc[i].to_dict()
                if self._contains_index:
                    r.pop("index")
                yield r
        self._csv.close()
        self._csv = None

    def __len__(self) -> int:
        self.require_loaded()
        return self._csv.shape[0]

    def load(self) -> None:
        if self._csv:
            return
        header_csv = _read_csv(self._file, chunksize=1)
        self._header = tuple(header_csv.get_chunk().columns)
        if self._header[0] == "index":
            self._header = self._header[1:]
            self._contains_index = True
        header_csv.close()
        self._csv = _read_csv(self._file, chunksize=self._chunk_size, low_memory=False)

    def save(self, file: str | _TextIO) -> None:
        self.require_loaded()
        csv = CSV(file, self._header)
        for row in self:
            csv.write_frame(*row.values())
        csv.close()

    def close(self) -> None:
        if self._csv:
            self._csv.close()


DEFAULT_HEADER: tuple[str, str, str, str, str, str, str, str, str, str, str, str] = (
    "t", "voltage", "speed", "front_wheel_speed", "rear_wheel_speed", "forward_acceleration", "lateral_acceleration",
    "mileage", "gps_valid", "gps_ground_speed", "latitude", "longitude"
)
DEFAULT_HEADER_FULL: tuple[str, str, str, str, str, str, str, str, str, str, str, str, str, str] = DEFAULT_HEADER + (
    "throttle", "brake"
)



from typing import TypeVar as _TypeVar, Any as _Any, override as _override, Literal as _Literal

from leads.context import Context
from leads.data import DataContainer
from leads.event import EventListener, Event, DataPushedEvent, UpdateEvent, SuspensionEvent, InterventionEvent, \
    InterventionExitEvent
from leads.plugin import Plugin

T = _TypeVar("T", bound=DataContainer)


class LEADS(Context[T]):
    def __init__(self, initial_data: T | None = None, data_seq_size: int = 100, num_laps_timed: int = 3) -> None:
        super().__init__(initial_data, data_seq_size, num_laps_timed)
        self._plugins: dict[tuple[str, ...], Plugin] = {}
        self._event_listener: EventListener = EventListener()

    def plugin(self, key: str | tuple[str, ...], plugin: Plugin | None = None) -> Plugin | None:
        if isinstance(key, str):
            key = key,
        if plugin is None:
            return self._plugins[key]
        self._plugins[key] = plugin
        plugin.on_load(self)

    def set_event_listener(self, event_listener: EventListener) -> None:
        event_listener.bind_chain(self._event_listener)
        self._event_listener = event_listener

    @_override
    def suspend(self, event: SuspensionEvent) -> None:
        self._event_listener.pre_suspend(event)

    def _acquire_data(self, name: str, *systems: str, mandatory: bool = True) -> _Any | None:
        try:
            return getattr(self.data(), name)
        except AttributeError:
            if mandatory:
                for system in systems:
                    self.suspend(SuspensionEvent(self, system, f"no data for `{name}`"))

    def _do_plugin_callback(self, method: _Literal["pre_push", "post_push", "pre_update", "post_update"]) -> None:
        for key, plugin in self._plugins.items():
            if plugin.enabled():
                getattr(plugin, method)(self, {d: self._acquire_data(d, *key) for d in plugin.required_data()})

    @_override
    def push(self, data: T) -> None:
        self._event_listener.pre_push(DataPushedEvent(self, data))
        self._do_plugin_callback("pre_push")
        super().push(data)
        self._do_plugin_callback("post_push")
        self._event_listener.post_push(DataPushedEvent(self, data))

    @_override
    def intervene(self, event: InterventionEvent) -> None:
        if isinstance(event, InterventionExitEvent):
            self._event_listener.post_intervene(event)
        else:
            self._event_listener.pre_intervene(event)

    @_override
    def update(self) -> None:
        self._do_plugin_callback("pre_update")
        self._event_listener.on_update(UpdateEvent(self))
        self._do_plugin_callback("post_update")

    @_override
    def left_indicator(self, left_indicator: bool | None = None, override: bool = False) -> bool | None:
        if (r := super().left_indicator(left_indicator, override)) is None:
            self._event_listener.left_indicator(Event("LEFT_INDICATOR", self), left_indicator)
        return r

    @_override
    def right_indicator(self, right_indicator: bool | None = None, override: bool = False) -> bool | None:
        if (r := super().right_indicator(right_indicator, override)) is None:
            self._event_listener.right_indicator(Event("RIGHT_INDICATOR", self), right_indicator)
        return r

    @_override
    def hazard(self, hazard: bool | None = None) -> bool | None:
        if (r := super().hazard(hazard)) is None:
            self._event_listener.hazard(Event("HAZARD", self), hazard)
        return r



from os.path import exists as _exists
from subprocess import run as _run, PIPE as _PIPE
from threading import Thread as _Thread

from leads import L as _L
from leads_gui.system import get_system_kernel as _get_system_kernel


def auto_path() -> str:
    return "%AppData%/frp" if _get_system_kernel() == "windows" else "/usr/local/frp"


def frpc_exists() -> bool:
    return _exists(f"{(path := auto_path())}/frpc") and _exists(f"{path}/frpc.toml")


def start_frpc() -> None:
    if not frpc_exists():
        raise FileNotFoundError("`frpc` not found")

    def wrapper() -> None:
        r = _run((f"{(path := auto_path())}/frpc", "-c", f"{path}/frpc.toml"), stdout=_PIPE, stderr=_PIPE)
        _L.error("`frpc` exits prematurely")
        _L.debug(f"Console output:\n{r.stdout.decode()}")

    _Thread(name="frpc", target=wrapper, daemon=True).start()



from typing import Self as _Self

from leads.os import _currentframe


class CallbackChain(object):
    def __init__(self, chain: _Self | None = None) -> None:
        self._chain: CallbackChain | None = chain

    def bind_chain(self, chain: _Self | None) -> None:
        self._chain = chain

    def super(self, *args, **kwargs) -> None:
        """
        Call the superior method if there is one.
        This must be called directly in the corresponding successor method.
        """
        if not self._chain:
            return
        cf = _currentframe().f_back
        while (cn := cf.f_code.co_name) == "super":
            cf = cf.f_back
        getattr(self._chain, cn)(*args, **kwargs)



from atexit import register as _register
from threading import Thread as _Thread
from time import sleep as _sleep

from sdl2 import AUDIO_S16 as _AUDIO_S16
from sdl2.sdlmixer import Mix_Init as _init, Mix_OpenAudioDevice as _open_audio_device, MIX_INIT_MP3 as _MIX_INIT_MP3, \
    Mix_LoadMUS as _load_music, Mix_PlayMusic as _play_music, Mix_Music as _Music, \
    Mix_GetError as _get_error, Mix_FreeMusic as _free_music, Mix_CloseAudio as _close_audio

from leads import L as _L
from leads_audio.system import _ASSETS_PATH


def _ensure(flag: int) -> None:
    if flag < 0:
        raise RuntimeError(_get_error().decode())


_init(_MIX_INIT_MP3)


def _try_open_audio_device() -> None:
    flag = -1
    while flag < 0:
        try:
            _ensure(flag := _open_audio_device(44100, _AUDIO_S16, 2, 2048, None, 1))
        except RuntimeError as e:
            _L.error(repr(e))
            _sleep(10)


_Thread(name="sdl initializer", target=_try_open_audio_device, daemon=True).start()

_register(_close_audio)


class _SoundEffect(object):
    def __init__(self, name: str) -> None:
        self._name: str = name
        self._source: _Music | None = None

    def load_source(self) -> _Music:
        if self._source is None:
            self._source = _load_music(f"{_ASSETS_PATH}/{self._name}.mp3".encode())
        return self._source

    def play(self) -> None:
        _ensure(_play_music(self.load_source(), 0))

    def stop(self) -> None:
        _ensure(_free_music(self.load_source()))


CONFIRM: _SoundEffect = _SoundEffect("confirm")
DIRECTION_INDICATOR_ON: _SoundEffect = _SoundEffect("direction-indicator-on")
DIRECTION_INDICATOR_OFF: _SoundEffect = _SoundEffect("direction-indicator-off")
WARNING: _SoundEffect = _SoundEffect("warning")



from importlib.util import find_spec as _find_spec

if not _find_spec("sdl2"):
    raise ImportError("Please install `PySDL2` and `pysdl2-dll` to run this module\n>>>pip install PySDL2 pysdl2-dll")

from leads_audio.prototype import *



from importlib.util import find_spec as _find_spec

if not _find_spec("matplotlib"):
    raise ImportError("Please install `matplotlib` to run this module\n>>>pip install matplotlib")

from datetime import datetime as _datetime
from typing import Any as _Any, Callable as _Callable, Sequence as _Sequence
from matplotlib.pyplot import figure as _figure, scatter as _scatter, show as _show, title as _title, \
    colorbar as _colorbar, bar as _bar, xticks as _xticks, legend as _legend, xlabel as _xlabel, ylabel as _ylabel

from leads.data import dlat2meters, dlon2meters, format_duration
from leads.data_persistence.analyzer.utils import time_invalid, speed_invalid, mileage_invalid, latitude_invalid, \
    longitude_invalid
from leads.data_persistence.core import CSVDataset, DEFAULT_HEADER
from .._computational import sqrt as _sqrt


class StaticProcessor(object):
    def __init__(self, dataset: CSVDataset) -> None:
        if DEFAULT_HEADER in dataset.read_header():
            raise KeyError("Your dataset must include the default header")
        self._dataset: CSVDataset = dataset

        # baking variables
        self._read_rows_count: int = 0
        self._valid_rows_count: int = 0
        self._invalid_rows: list[int] = []
        self._start_time: int | None = None
        self._end_time: int | None = None
        self._duration: int | None = None
        self._min_speed: float | None = None
        self._max_speed: float | None = None
        self._avg_speed: float | None = None
        self._start_mileage: float | None = None
        self._end_mileage: float | None = None
        self._distance: float | None = None
        self._gps_valid_count: int = 0
        self._gps_invalid_rows: list[int] = []
        self._min_lat: float | None = None
        self._min_lon: float | None = None

        # process variables
        self._laps: list[tuple[int, int, int, float, float]] = []
        self._max_lap_duration: int | None = None
        self._max_lap_distance: float | None = None
        self._max_lap_avg_speed: float | None = None

        # unit variables (not reusable)
        self._lap_start: int | None = None
        self._lap_start_time: int | None = None
        self._lap_start_mileage: float | None = None
        self._lap_x: list[float] = []
        self._lap_y: list[float] = []
        self._lap_d: list[float] = []
        self._max_lap_x: float | None = None
        self._max_lap_y: float | None = None

    def dataset(self) -> CSVDataset:
        return self._dataset

    def bake(self) -> None:
        """
        Prepare the prerequisites for `process()`.
        """

        def unit(row: dict[str, _Any], i: int) -> None:
            self._read_rows_count += 1
            t = int(row["t"])
            speed = row["speed"]
            mileage = row["mileage"]
            if time_invalid(t) or speed_invalid(
                    speed) or mileage_invalid(mileage):
                self._invalid_rows.append(i)
                return
            if self._start_time is None:
                self._start_time = t
            self._end_time = t
            if self._min_speed is None or speed < self._min_speed:
                self._min_speed = speed
            if self._max_speed is None or speed > self._max_speed:
                self._max_speed = speed
            if self._start_mileage is None:
                self._start_mileage = mileage
            self._end_mileage = mileage
            lat = row["latitude"]
            lon = row["longitude"]
            if not row["gps_valid"] or latitude_invalid(lat) or longitude_invalid(lon):
                self._gps_invalid_rows.append(i)
            else:
                if self._min_lat is None or lat < self._min_lat:
                    self._min_lat = lat
                if self._min_lon is None or lon < self._min_lon:
                    self._min_lon = lon
                self._gps_valid_count += 1
            self._valid_rows_count += 1

        self.foreach(unit, False)
        if self._valid_rows_count == 0:
            raise LookupError("Failed to bake")
        self._duration = self._end_time - self._start_time
        self._distance = self._end_mileage - self._start_mileage
        self._avg_speed = 3600000 * self._distance / self._duration

    @staticmethod
    def _hide_others(seq: _Sequence[_Any], limit: int) -> str:
        return f"[{", ".join(map(str, seq[:limit]))}, and {diff} others]" if (diff := len(seq) - limit) > 0 else str(
            seq)

    def baking_results(self) -> tuple[str, str, str, str, str, str, str, str, str, str, str, str]:
        """
        Get the results of the baking process.
        :return: the results in sentences
        """
        if self._read_rows_count == 0:
            raise LookupError("Not baked")
        if self._valid_rows_count == 0:
            raise LookupError(
                f"Failed to baked {self._read_rows_count - self._valid_rows_count} / {self._read_rows_count} rows")
        return (
            f"Baked {self._valid_rows_count} / {self._read_rows_count} ROWS",
            f"Baking Rate: {100 * self._valid_rows_count / self._read_rows_count:.2f}%",
            f"Skipped Rows: {StaticProcessor._hide_others(self._invalid_rows, 5)}",
            f"Start Time: {_datetime.fromtimestamp(self._start_time * .001).strftime("%Y-%m-%d %H:%M:%S")}",
            f"End Time: {_datetime.fromtimestamp(self._end_time * .001).strftime("%Y-%m-%d %H:%M:%S")}",
            f"Duration: {format_duration(self._duration * .001)}",
            f"Distance: {self._distance:.2f} KM",
            f"v\u2098\u1D62\u2099: {self._min_speed:.2f} KM / H",
            f"v\u2098\u2090\u2093: {self._max_speed:.2f} KM / H",
            f"v\u2090\u1D65\u1D4D: {self._avg_speed:.2f} KM / H",
            f"GPS Hit Rate: {100 * self._gps_valid_count / self._valid_rows_count:.2f}%",
            f"GPS Skipped Rows: {StaticProcessor._hide_others(self._gps_invalid_rows, 5)}"
        )

    def erase_unit_cache(self) -> None:
        self._lap_start = None
        self._lap_start_time = None
        self._lap_start_mileage = None
        self._lap_x.clear()
        self._lap_y.clear()
        self._lap_d.clear()

    def foreach(self, do: _Callable[[dict[str, _Any], int], None], skip_invalid_rows: bool = True,
                skip_gps_invalid_rows: bool = False) -> None:
        self.erase_unit_cache()
        i = -1
        for row in self._dataset:
            i += 1
            if skip_invalid_rows and i in self._invalid_rows or skip_gps_invalid_rows and i in self._gps_invalid_rows:
                continue
            do(row, i)

    def process(self, lap_time_assertions: _Sequence[float] | None = None, vehicle_hit_box: float = 3,
                min_lap_time: float = 30) -> None:
        """
        Split the laps.
        :param lap_time_assertions: the manually timed laps in seconds
        :param vehicle_hit_box: the vehicle hit box in meters
        :param min_lap_time: the minimum lap time in seconds
        :return:
        """
        asserted = lap_time_assertions is not None

        def shared_pre(row: dict[str, _Any], i: int) -> tuple[int, float, float]:
            if self._lap_start is None:
                self._lap_start = i
            t = int(row["t"])
            if self._lap_start_time is None:
                self._lap_start_time = t
            mileage = row["mileage"]
            if self._lap_start_mileage is None:
                self._lap_start_mileage = mileage
            return (dt := t - self._lap_start_time), (
                ds := mileage - self._lap_start_mileage), 3600000 * ds / dt if dt else 0

        def shared_post(duration: float, distance: float, avg_speed: float) -> None:
            if self._max_lap_duration is None or duration > self._max_lap_duration:
                self._max_lap_duration = duration
            if self._max_lap_distance is None or distance > self._max_lap_distance:
                self._max_lap_distance = distance
            if self._max_lap_avg_speed is None or avg_speed > self._max_lap_avg_speed:
                self._max_lap_avg_speed = avg_speed

        def asserted_unit(row: dict[str, _Any], i: int) -> None:
            duration, distance, avg_speed = shared_pre(row, i)
            next_lap_index = len(self._laps)
            if next_lap_index < len(lap_time_assertions) and duration >= lap_time_assertions[next_lap_index] * 1000:
                shared_post(duration, distance, avg_speed)
                self._laps.append((self._lap_start, i, duration, distance, avg_speed))
                self.erase_unit_cache()

        path = []

        def unit(row: dict[str, _Any], i: int) -> None:
            lat = row["latitude"]
            lon = row["longitude"]
            p = (round(dlat2meters(lat - self._min_lat) / vehicle_hit_box),
                 round(dlon2meters(lon - self._min_lon, lat) / vehicle_hit_box))
            try:
                index = path.index(p)
            except ValueError:
                index = -1
            duration, distance, avg_speed = shared_pre(row, i)
            if (0 < index < .5 * len(path) and self._lap_start is not None and duration >= min_lap_time * 1000 and
                    distance * 2000 > vehicle_hit_box):
                shared_post(duration, distance, avg_speed)
                self._laps.append((self._lap_start, i, duration, distance, avg_speed))
                path.clear()
                self.erase_unit_cache()
            else:
                path.append(p)

        self.foreach(asserted_unit if asserted else unit, True, not asserted)

    def suggest_on_lap(self, lap_index: int) -> tuple[str, str]:
        a, b, duration, distance, avg_speed = self._laps[lap_index]
        d = self._avg_speed - avg_speed
        return (
            f"Lap {lap_index + 1} lasts for {format_duration(duration * .001)}",
            f"{abs(d):.2f} KM / H {"slower" if d < 0 else "faster"} than average"
        )

    def num_laps(self) -> int:
        return len(self._laps)

    def results(self) -> tuple[str, str]:
        """
        Get the results of the processor.
        :return: the results in sentences
        """
        return (
            f"Number of Laps Detected: {len(self._laps)}",
            f"Call `draw_lap()` for further information."
        )

    def close(self) -> None:
        self._dataset.close()

    def draw_lap(self, lap_index: int = -1) -> None:
        if lap_index < 0:
            for i in range(len(self._laps)):
                self.draw_lap(i)
        if lap_index >= len(self._laps):
            raise IndexError("Lap index out of range")

        def unit(row: dict[str, _Any], index: int) -> None:
            if index < (lap := self._laps[lap_index])[0] or index > lap[1]:
                return
            t = int(row["t"])
            if self._lap_start_time is None:
                self._lap_start_time = t
            self._lap_end_time = t
            lat = row["latitude"]
            lon = row["longitude"]
            self._lap_d.append(row["speed"])
            self._lap_x.append(x := dlon2meters(lon - self._min_lon, lat))
            self._lap_y.append(y := dlat2meters(lat - self._min_lat))
            if self._max_lap_x is None or x > self._max_lap_x:
                self._max_lap_x = x
            if self._max_lap_y is None or y > self._max_lap_y:
                self._max_lap_y = y

        self.foreach(unit, True, True)
        far = max(self._max_lap_x, self._max_lap_y)
        self._lap_x.append(far)
        self._lap_y.append(far)
        self._lap_d.append(self._max_speed)
        _figure(figsize=(6, 5))
        _title(f"Lap {lap_index + 1} ({self._laps[lap_index][3]:.2f} KM @ {format_duration(
            self._laps[lap_index][2] * .001)})")
        _scatter(self._lap_x, self._lap_y, c=self._lap_d, cmap="hot_r")
        _xlabel("X (M)")
        _ylabel("Y (M)")
        cb = _colorbar()
        cb.set_label("Speed (KM / H)")
        cb.ax.hlines(self._laps[lap_index][4], 0, 1)
        _show()

    def draw_comparison_of_laps(self, width: float = .3) -> None:
        durations = []
        x0 = []
        distances = []
        x1 = []
        avg_speeds = []
        x2 = []
        x_ticks = []
        i = 1
        for lap in self._laps:
            durations.append(lap[2] / self._max_lap_duration)
            x0.append(i)
            distances.append(lap[3] / self._max_lap_distance)
            x1.append(i + width)
            avg_speeds.append(lap[4] / self._max_lap_avg_speed)
            x2.append(i + 2 * width)
            x_ticks.append(f"L{i}")
            i += 1
        _figure(figsize=(5 * _sqrt(len(self._laps)), 5))
        _bar(x0, durations, width, label="Duration")
        _bar(x1, distances, width, label="Distance")
        _bar(x2, avg_speeds, width, label="Average Speed")
        _xticks(x1, x_ticks)
        _legend()
        _xlabel("Lap")
        _ylabel("Proportion (% / max)")
        _show()



from typing import Any as _Any, override as _override

from leads.constant import ESCMode
from leads.context import Context
from leads.registry import require_context


class Plugin(object):
    def __init__(self, required_data: tuple[str, ...] = ()) -> None:
        """
        :param required_data: required data entries
        """
        super().__init__()
        self._required_data: tuple[str, ...] = required_data
        self.state: dict[str, _Any] = {}
        self._enabled: bool = True

    def __getitem__(self, key: str) -> _Any:
        return self.state[key]

    def __setitem__(self, key: str, value: _Any) -> None:
        self.state[key] = value

    def enabled(self, enabled: bool | None = None) -> bool | None:
        if enabled is None:
            return self._enabled
        self._enabled = enabled

    def required_data(self) -> tuple[str, ...]:
        return self._required_data

    def on_load(self, context: Context) -> None: ...

    def pre_push(self, context: Context, kwargs: dict[str, _Any]) -> None:
        """
        Note that the new data at this point is not available yet.
        :param context: target context
        :param kwargs: {required data: its value}
        """
        ...

    def post_push(self, context: Context, kwargs: dict[str, _Any]) -> None:
        """
        :param context: target context
        :param kwargs: {required data: its value}
        """
        ...

    def pre_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        """
        :param context: target context
        :param kwargs: {required data: its value}
        """
        ...

    def post_update(self, context: Context, kwargs: dict[str, _Any]) -> None:
        """
        :param context: target context
        :param kwargs: {required data: its value}
        """
        ...


class ESCPlugin(Plugin):
    @_override
    def enabled(self, enabled: bool | None = None) -> bool | None:
        if enabled is None:
            return require_context().esc_mode() != ESCMode.OFF and super().enabled()
        super().enabled(enabled)

    @staticmethod
    def adjudicate(d: float, base: float, absolute: float, percentage: float) -> bool:
        return d > 0 and ((absolute and d > absolute) or (percentage and d > base * percentage))



from importlib.util import find_spec as _find_spec

if not _find_spec("fastapi"):
    raise ImportError("Please install `fastapi` to run this module\n>>>pip install \"fastapi[all]\"")
if not _find_spec("uvicorn"):
    raise ImportError("Please install `uvicorn` to run this module\n>>>pip install \"fastapi[all]\"")



from random import randint as _randint
from time import time as _time
from typing import override as _override

from numpy import sin as _sin, pi as _pi

from leads import Controller as _Controller, DataContainer as _DataContainer


class _EmulatedController(_Controller):
    def __init__(self,
                 minimum: int = 30,
                 maximum: int = 40,
                 skid_possibility: float = .1) -> None:
        super().__init__()
        self.minimum: int = minimum
        self.maximum: int = maximum
        self.skid_possibility: float = skid_possibility
        self._last_speed: float = 0
        self._last_time: float = _time()

    def generate_rear_wheel_speed(self, front_wheel_speed: float) -> float:
        return front_wheel_speed if self.skid_possibility <= 0 else front_wheel_speed + int(_randint(
            -int(1 / self.skid_possibility), int(1 / self.skid_possibility)) * self.skid_possibility)

    def generate_forward_acceleration(self, speed: float) -> float:
        t = _time()
        try:
            return (speed - self._last_speed) / 3.6 / (t - self._last_time)
        finally:
            self._last_speed = speed
            self._last_time = t


class RandomController(_EmulatedController):
    @_override
    def read(self) -> _DataContainer:
        return _DataContainer(voltage=48.0,
                              speed=(fws := _randint(self.minimum, self.maximum)),
                              front_wheel_speed=fws,
                              rear_wheel_speed=self.generate_rear_wheel_speed(fws),
                              forward_acceleration=self.generate_forward_acceleration(fws),
                              gps_valid=True,
                              gps_ground_speed=fws,
                              latitude=_randint(4315, 4415) / 100,
                              longitude=-_randint(7888, 7988) / 100)


class SinController(_EmulatedController):
    def __init__(self,
                 minimum: int = 30,
                 maximum: int = 40,
                 skid_possibility: float = .1,
                 acceleration: float = .01) -> None:
        super().__init__(minimum, maximum, skid_possibility)
        self.acceleration: float = acceleration
        self.magnitude: int = int((maximum - minimum) * .5)
        self.counter: float = 0

    @_override
    def read(self) -> _DataContainer:
        try:
            return _DataContainer(voltage=48.0,
                                  speed=(fws := (_sin(self.counter) * self.magnitude + self.magnitude)),
                                  front_wheel_speed=fws,
                                  rear_wheel_speed=self.generate_rear_wheel_speed(fws),
                                  forward_acceleration=self.generate_forward_acceleration(fws),
                                  gps_valid=True,
                                  gps_ground_speed=fws,
                                  latitude=_randint(4315, 4415) / 100,
                                  longitude=-_randint(7888, 7988) / 100)
        finally:
            self.counter = (self.counter + self.acceleration) % (2 * _pi)



MAIN_CONTROLLER: str = "__main__"

LEFT_FRONT_WHEEL_SPEED_SENSOR: str = "wss_lf"
RIGHT_FRONT_WHEEL_SPEED_SENSOR: str = "wss_rf"
CENTER_REAR_WHEEL_SPEED_SENSOR: str = "wss_cr"
LEFT_REAR_WHEEL_SPEED_SENSOR: str = "wss_lr"
RIGHT_REAR_WHEEL_SPEED_SENSOR: str = "wss_rr"
ACCELEROMETER: str = "accl"
ODOMETER: str = "odm"
VOLTAGE_SENSOR: str = "vot"
THROTTLE_PEDAL: str = "tpd"
BRAKE_PEDAL: str = "bpd"
MOTOR_CONTROLLER: str = "mc"
BRAKE_CONTROLLER: str = "bc"
GPS_RECEIVER: str = "gps"
LEFT_INDICATOR: str = "lindi"
RIGHT_INDICATOR: str = "rindi"
